import warnings
from typing import Dict, Union, Optional

import numpy as np
from tqdm import tqdm
import sklearn.metrics as met

from peyes._DataModels.EventLabelEnum import EventLabelEnum, EventLabelSequenceType
from peyes._DataModels.UnparsedEventLabel import UnparsedEventLabelType, UnparsedEventLabelSequenceType

import peyes._utils.constants as cnst
from peyes._utils.event_utils import parse_label
from peyes._utils.metric_utils import normalized_levenshtein_distance as _nld
from peyes._utils.metric_utils import dprime_and_criterion as _dprime_and_criterion

_GLOBAL_METRICS = {
    cnst.ACCURACY_STR, cnst.BALANCED_ACCURACY_STR, cnst.COHENS_KAPPA_STR, cnst.MCC_STR, cnst.COMPLEMENT_NLD_STR,
}
_SDT_METRICS = {
    cnst.RECALL_STR, cnst.PRECISION_STR, cnst.F1_STR, cnst.D_PRIME_STR, cnst.CRITERION_STR
}


def calculate(
        ground_truth: UnparsedEventLabelSequenceType,
        prediction: UnparsedEventLabelSequenceType,
        *metrics: str,
        pos_labels: Optional[Union[UnparsedEventLabelType, UnparsedEventLabelSequenceType]] = None,
        verbose: bool = False,
        **kwargs,
) -> Union[float, Dict[str, float]]:
    """
    Calculate the specified metrics between the ground truth and prediction sequences.
    :param ground_truth: sequence of ground truth labels
    :param prediction: sequence of predicted labels
    :param metrics: the metrics to calculate. Supported metrics are:
        - "accuracy"
        - "balanced_accuracy"
        - "cohen's_kappa"
        - "mcc" - mathew's correlation coefficient
        - "complement_nld" - computed the complement to normalized Levenshtein distance (1 - WER)
        - "recall"
        - "precision"
        - "f1"
        - "d_prime" or "d'"
        - "criterion"
    :param pos_labels: the positive labels to consider for recall, precision, f1-score, d-prime, and criterion
    :param verbose: if True, display a progress bar

    :keyword average: the averaging strategy for recall, precision, and f1-score. default is "weighted"
    :keyword correction: the correction strategy for d-prime and criterion. default is "loglinear"
        See information on correction methods at https://stats.stackexchange.com/a/134802/288290.
        See implementation details at https://lindeloev.net/calculating-d-in-python-and-php/.

    :return: the calculated metric(s) as a single float (if only one metric is specified) or a dictionary of metric
        names to values
    """
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=UserWarning)
        assert len(ground_truth) == len(prediction), "Ground Truth and Prediction must have the same length."
        ground_truth = [parse_label(label) for label in ground_truth]
        prediction = [parse_label(label) for label in prediction]
        results: Dict[str, float] = {}
        for metric in tqdm(metrics, desc="Calculating Metrics", disable=not verbose):
            metric_lower = metric.lower().strip().replace(" ", "_").replace("-", "_").removesuffix("_score")
            if metric_lower in _GLOBAL_METRICS:
                results[metric] = _calculate_global_metrics(ground_truth, prediction, metric_lower)
            elif metric_lower in _SDT_METRICS:
                average = kwargs.get("average", "weighted").lower().strip()
                correction = kwargs.get("correction", "loglinear").lower().strip()
                results[metric] = _calculate_sdt_metrics(
                    ground_truth, prediction, metric_lower, pos_labels, average, correction
                )
            else:
                raise NotImplementedError(f"Unknown metric:\t{metric}")
        if len(results) == 1:
            return next(iter(results.values()))
        return results


def _calculate_global_metrics(
        ground_truth: EventLabelSequenceType,
        prediction: EventLabelSequenceType,
        metric: str,
) -> float:
    if metric == cnst.ACCURACY_STR:
        return met.accuracy_score(ground_truth, prediction)
    if metric == cnst.BALANCED_ACCURACY_STR:
        return met.balanced_accuracy_score(ground_truth, prediction)
    if metric == cnst.MCC_STR or metric == "mathew's_correlation":
        return met.matthews_corrcoef(ground_truth, prediction)
    if metric == cnst.COHENS_KAPPA_STR or metric == "cohen_kappa":
        return met.cohen_kappa_score(ground_truth, prediction)
    if metric == cnst.COMPLEMENT_NLD_STR or metric == "1_nld":
        nld = _nld(ground_truth, prediction)
        return 1 - nld
    raise NotImplementedError(f"Unknown metric:\t{metric}")


def _calculate_sdt_metrics(
        ground_truth: EventLabelSequenceType,
        prediction: EventLabelSequenceType,
        metric: str,
        pos_labels: Optional[Union[UnparsedEventLabelType, UnparsedEventLabelSequenceType]] = None,
        average: str = "weighted",
        correction: str = "loglinear",
) -> float:
    average = average.lower().strip()
    if pos_labels is None:
        pos_labels = [l for l in EventLabelEnum]
    elif isinstance(pos_labels, UnparsedEventLabelType):
        pos_labels = [parse_label(pos_labels)]
    else:
        pos_labels = [parse_label(l) for l in pos_labels]
    if metric == cnst.RECALL_STR:
        return met.recall_score(ground_truth, prediction, labels=pos_labels, average=average, zero_division=np.nan)
    if metric == cnst.PRECISION_STR:
        return met.precision_score(ground_truth, prediction, labels=pos_labels, average=average, zero_division=np.nan)
    if metric == cnst.F1_STR:
        return met.f1_score(ground_truth, prediction, labels=pos_labels, average=average, zero_division=np.nan)
    if metric in {"dprime", "d'", cnst.D_PRIME_STR, cnst.CRITERION_STR}:
        p = np.sum([1 for label in ground_truth if label in pos_labels])
        n = len(ground_truth) - p
        pp = np.sum([1 for label in prediction if label in pos_labels])
        tp = np.sum([1 for gt, pred in zip(ground_truth, prediction) if pred == gt and gt in pos_labels])
        dprime, crit = _dprime_and_criterion(p, n, pp, tp, correction)
        return dprime if metric == cnst.D_PRIME_STR else crit
    raise NotImplementedError(f"Unknown metric:\t{metric}")



import warnings
from abc import ABC
from typing import final, List, Optional, Sequence

import pandas as pd

import peyes._DataModels.config as cnfg
from peyes._utils.pixel_utils import *
from peyes._utils.vector_utils import get_chunk_indices
from peyes._DataModels.EventLabelEnum import EventLabelEnum


class BaseEvent(ABC):

    _LABEL: EventLabelEnum
    _OUTLIER_REASONS_STR = "outlier_reasons"

    def __init__(
            self,
            t: np.ndarray,
            x: np.ndarray = None,
            y: np.ndarray = None,
            pupil: np.ndarray = None,
            viewer_distance: float = cnfg.VIEWER_DISTANCE,
            pixel_size: float = cnfg.SCREEN_MONITOR[cnst.PIXEL_SIZE_STR],
    ):
        _x = x if x is not None else np.full_like(t, np.nan, dtype=float)
        _y = y if y is not None else np.full_like(t, np.nan, dtype=float)
        _pupil = pupil if pupil is not None else np.full_like(t, np.nan, dtype=float)
        assert len(t) == len(_x) == len(_y) == len(_pupil), "t, x, y, and pupil must have the same length"
        assert np.isnan(viewer_distance) or viewer_distance > 0, "viewer_distance must be a positive number"
        assert np.isnan(pixel_size) or pixel_size > 0, "pixel_size must be a positive number"
        samples = np.vstack([t, _x, _y, _pupil]).T
        samples = samples[samples[:, 0].argsort()]  # sort by time
        self._t = samples[:, 0]
        self._x = samples[:, 1]
        self._y = samples[:, 2]
        self._pupil = samples[:, 3]
        self._viewer_distance = viewer_distance
        self._pixel_size = pixel_size

    @staticmethod
    @final
    def make(
            label: EventLabelEnum,
            t: np.ndarray,
            x: np.ndarray = None,
            y: np.ndarray = None,
            pupil: np.ndarray = None,
            viewer_distance: float = cnfg.VIEWER_DISTANCE,
            pixel_size: float = cnfg.SCREEN_MONITOR[cnst.PIXEL_SIZE_STR],
    ) -> Optional["BaseEvent"]:
        """  Creates a single Event from the given data.  """
        if label == EventLabelEnum.UNDEFINED:
            return None
        if label == EventLabelEnum.FIXATION:
            return FixationEvent(t, x, y, pupil, viewer_distance, pixel_size)
        if label == EventLabelEnum.SACCADE:
            return SaccadeEvent(t, x, y, pupil, viewer_distance, pixel_size)
        if label == EventLabelEnum.PSO:
            return PSOEvent(t, x, y, pupil, viewer_distance, pixel_size)
        if label == EventLabelEnum.SMOOTH_PURSUIT:
            return SmoothPursuitEvent(t, x, y, pupil, viewer_distance, pixel_size)
        if label == EventLabelEnum.BLINK:
            return BlinkEvent(t, x, y, pupil, viewer_distance, pixel_size)
        raise ValueError(f"Invalid event type: {label}")

    @staticmethod
    @final
    def make_multiple(
            labels: np.ndarray,
            t: np.ndarray,
            x: np.ndarray = None,
            y: np.ndarray = None,
            pupil: np.ndarray = None,
            viewer_distance: float = cnfg.VIEWER_DISTANCE,
            pixel_size: float = cnfg.SCREEN_MONITOR[cnst.PIXEL_SIZE_STR],
    ) -> list["BaseEvent"]:
        if len(labels) != len(t):
            raise ValueError("Length of `labels` and `t` must be the same")
        same_label_chunk_idxs = get_chunk_indices(labels)
        events = []
        for idx_group in same_label_chunk_idxs:
            event_label = EventLabelEnum(labels[idx_group[0]])
            if event_label == EventLabelEnum.UNDEFINED:
                # ignore undefined samples
                continue
            new_event = BaseEvent.make(
                label=event_label,
                t=t[idx_group],
                x=x[idx_group] if x is not None else None,
                y=y[idx_group] if y is not None else None,
                pupil=pupil[idx_group] if pupil is not None else None,
                viewer_distance=viewer_distance,
                pixel_size=pixel_size
            )
            if new_event is not None:
                events.append(new_event)
        return events

    @final
    def velocities(self, unit: str = 'px') -> np.ndarray:
        """
        Calculates the velocity (distance per second) of each sample within the event.
        :param unit: determines the nominator units
        :return: np.ndarray of velocities (units are one of px/s, deg/s, rad/s)
        """
        px_velocities = calculate_velocities(self._x, self._y, self._t)
        unit = unit.lower()
        if unit in {"px", "pixel", "pixels", "px/sec"}:
            return px_velocities
        px_to_visual_angle_vec = np.vectorize(pixels_to_visual_angle)
        if unit in {"deg", "degree", "degrees", "deg/sec"}:
            return px_to_visual_angle_vec(px_velocities, self._viewer_distance, self._pixel_size)
        if unit in {"rad", "radian", "radians", "rad/sec"}:
            return px_to_visual_angle_vec(px_velocities, self._viewer_distance, self._pixel_size, use_radians=True)
        raise ValueError(f"unit must be one of `px`, `deg`, or `rad`, not `{unit}`")

    def get_outlier_reasons(self) -> List[str]:
        reasons = []
        if self.duration < cnfg.EVENT_MAPPING[self._LABEL][cnst.MIN_DURATION_STR]:
            reasons.append(cnst.MIN_DURATION_STR)
        if self.duration > cnfg.EVENT_MAPPING[self._LABEL][cnst.MAX_DURATION_STR]:
            reasons.append(cnst.MAX_DURATION_STR)
        if np.any(self._x < 0) or np.any(self._y < 0):
            reasons.append("negative_pixel")
        if (
                np.any(self._x > cnfg.SCREEN_MONITOR[cnst.RESOLUTION_STR][0]) or
                np.any(self._y > cnfg.SCREEN_MONITOR[cnst.RESOLUTION_STR][1])
        ):
            reasons.append("pixel_outside_screen")
        # TODO: check min, max velocity, acceleration, dispersion, etc.
        return reasons

    def summary(self) -> pd.Series:
        with warnings.catch_warnings():
            # ignore warnings about NaN values in the event summary
            warnings.simplefilter("ignore", category=RuntimeWarning)
            d = {
                f"{cnst.EVENT_STR}_{cnst.TYPE_STR}": self.label.name,
                cnst.LABEL_STR: self.label.value,
                cnst.START_TIME_STR: self.start_time,
                cnst.END_TIME_STR: self.end_time,
                cnst.DURATION_STR: self.duration,
                cnst.DISTANCE_STR: self.distance,
                cnst.AMPLITUDE_STR: self.amplitude,
                cnst.AZIMUTH_STR: self.azimuth,
                cnst.PEAK_VELOCITY_STR: self.peak_velocity,
                cnst.MEDIAN_VELOCITY_STR: self.median_velocity,
                cnst.MIN_VELOCITY_STR: self.min_velocity,
                cnst.CUMULATIVE_DISTANCE_STR: self.cumulative_distance,
                cnst.CUMULATIVE_AMPLITUDE_STR: self.cumulative_amplitude,
                cnst.CENTER_PIXEL_STR: self.center_pixel,
                cnst.PIXEL_STD_STR: self.pixel_std,
                cnst.DISPERSION_STR: self.dispersion,
                cnst.ELLIPSE_AREA_STR: self.ellipse_area,
                cnst.IS_OUTLIER_STR: self.is_outlier,
                self._OUTLIER_REASONS_STR: self.get_outlier_reasons()
            }
            return pd.Series(d)

    @final
    def time_overlap(self, other: "BaseEvent", normalize: bool = True) -> float:
        """  Calculates the time overlap between this event and another event, in milliseconds.  """
        if self.duration == 0 or other.duration == 0:
            return 0
        start_time = max(self.start_time, other.start_time)
        end_time = min(self.end_time, other.end_time)
        total_overlap = max([0, end_time - start_time])
        if normalize:
            return total_overlap / self.duration
        return total_overlap

    @final
    def time_iou(self, other: "BaseEvent") -> float:
        """
        Calculates the intersection-over-union (IoU) between times of this event and another event (unitless).
        See Startsev & Zemblys (2023) for more information.
        """
        min_start, max_start = min(self.start_time, other.start_time), max(self.start_time, other.start_time)
        min_end, max_end = min(self.end_time, other.end_time), max(self.end_time, other.end_time)
        min_duration = max([0, min_end - max_start])
        max_duration = max([0, max_end - min_start])
        if max_duration == 0:
            return np.nan
        return min_duration / max_duration

    @final
    def time_l2(self, other: "BaseEvent") -> float:
        """
        Calculates the l2-norm between the onset- and offset-differences of this event and another event.
        See Kothari et al. (2020) for more details.
        """
        start_diff = self.start_time - other.start_time
        end_diff = self.end_time - other.end_time
        l2 = np.linalg.norm([start_diff, end_diff])
        return float(l2)

    @final
    def center_distance(self, other: "BaseEvent") -> float:
        """  Calculates the Euclidean distance between the center pixel of this event and another event (px units).  """
        x1, y1 = self.center_pixel
        x2, y2 = other.center_pixel
        return float(np.linalg.norm([x1 - x2, y1 - y2]))

    @final
    @property
    def label(self) -> EventLabelEnum:
        return self.__class__._LABEL

    @final
    @property
    def num_samples(self) -> int:
        return len(self._t)

    @final
    @property
    def is_outlier(self) -> bool:
        return len(self.get_outlier_reasons()) > 0

    @final
    @property
    def start_time(self) -> float:
        return float(self._t[0])

    @final
    @property
    def start_pixel(self) -> Tuple[float, float]:
        return float(self._x[0]), float(self._y[0])

    @final
    @property
    def end_time(self) -> float:
        return float(self._t[-1])

    @final
    @property
    def end_pixel(self) -> Tuple[float, float]:
        return float(self._x[-1]), float(self._y[-1])

    @final
    @property
    def duration(self) -> float:
        return float(self.end_time - self.start_time)

    @final
    @property
    def distance(self) -> float:
        """  Euclidean distance between the start and end pixels of the event (pixel units)  """
        start_x, start_y = self.start_pixel
        end_x, end_y = self.end_pixel
        return float(np.linalg.norm([end_x - start_x, end_y - start_y]))

    @final
    @property
    def amplitude(self) -> float:
        """  Euclidean distance between the start and end pixels of the event (visual degree units)  """
        return pixels_to_visual_angle(self.distance, self._viewer_distance, self._pixel_size)

    @final
    @property
    def azimuth(self) -> float:
        """  Angle between the start and end pixels of the event (degrees)"""
        return calculate_azimuth(self.start_pixel, self.end_pixel, use_radians=False)

    @final
    @property
    def cumulative_distance(self) -> float:
        """  Cumulative distance traveled during the event (pixel units)  """
        return np.sum(np.linalg.norm(np.diff(np.column_stack((self._x, self._y)), axis=0), axis=1))

    @final
    @property
    def cumulative_amplitude(self) -> float:
        """  Cumulative distance traveled during the event (visual degree units)  """
        return pixels_to_visual_angle(self.cumulative_distance, self._viewer_distance, self._pixel_size)

    @final
    @property
    def center_pixel(self) -> Tuple[float, float]:
        """  Returns the mean coordinates of the event on the X,Y axes  """
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)
            x_mean = float(np.nanmean(self._x))
            y_mean = float(np.nanmean(self._y))
            return x_mean, y_mean

    @final
    @property
    def pixel_std(self) -> Tuple[float, float]:
        """  Returns the standard deviation of the event (in pixel units)  """
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)
            x_std = float(np.nanstd(self._x))
            y_std = float(np.nanstd(self._y))
            return x_std, y_std

    @final
    @property
    def top_pixel(self) -> Tuple[float, float]:
        """  Returns the top pixel of the event (assuming the screen's top-left corner is (0,0))  """
        min_y_idx = np.argmin(self._y)
        return float(self._x[min_y_idx]), float(self._y[min_y_idx])

    @final
    @property
    def bottom_pixel(self) -> Tuple[float, float]:
        """  Returns the bottom pixel of the event (assuming the screen's top-left corner is (0,0))  """
        max_y_idx = np.argmax(self._y)
        return float(self._x[max_y_idx]), float(self._y[max_y_idx])

    @final
    @property
    def left_pixel(self) -> Tuple[float, float]:
        """  Returns the leftmost pixel of the event (assuming the screen's top-left corner is (0,0))  """
        min_x_idx = np.argmin(self._x)
        return float(self._x[min_x_idx]), float(self._y[min_x_idx])

    @final
    @property
    def right_pixel(self) -> Tuple[float, float]:
        """  Returns the rightmost pixel of the event (assuming the screen's top-left corner is (0,0))  """
        max_x_idx = np.argmax(self._x)
        return float(self._x[max_x_idx]), float(self._y[max_x_idx])

    @final
    @property
    def x_dispersion(self) -> float:
        """  Returns the horizontal dispersion of the event (visual degree units)  """
        return pixels_to_visual_angle(np.nanmax(self._x) - np.nanmin(self._x), self._viewer_distance, self._pixel_size)

    @final
    @property
    def y_dispersion(self) -> float:
        """  Returns the vertical dispersion of the event (visual degree units)  """
        return pixels_to_visual_angle(np.nanmax(self._y) - np.nanmin(self._y), self._viewer_distance, self._pixel_size)

    @final
    @property
    def dispersion(self) -> float:
        """  Returns the dispersion of the event - sum of horizontal and vertical dispersion (visual degree units)  """
        return self.x_dispersion + self.y_dispersion

    @final
    @property
    def ellipse_area(self) -> float:
        """  Returns the area of the ellipse that encloses the event (visual degree units)  """
        x_radius = self.x_dispersion / 2
        y_radius = self.y_dispersion / 2
        return np.pi * x_radius * y_radius

    @final
    @property
    def peak_velocity(self) -> float:
        """  Returns the maximum velocity during the event (visual degree / second)  """
        return np.nanmax(self.velocities(unit='deg'))

    @final
    @property
    def median_velocity(self) -> float:
        """  Returns the median velocity during the event (visual degree / second)  """
        return float(np.nanmedian(self.velocities(unit='deg')))

    @final
    @property
    def min_velocity(self) -> float:
        """  Returns the minimum velocity during the event (visual degree / second)  """
        return np.nanmin(self.velocities(unit='deg'))

    def __hash__(self):
        return hash((
            self.label,
            self._t.tobytes(),
            self._x.tobytes(),
            self._y.tobytes(),
            self._pupil.tobytes(),
            self._viewer_distance,
            self._pixel_size
        ))

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        if not self.label == other.label:
            return False
        if not np.array_equal(self._t, other._t, equal_nan=True):
            return False
        if not np.array_equal(self._x, other._x, equal_nan=True):
            return False
        if not np.array_equal(self._y, other._y, equal_nan=True):
            return False
        if not np.array_equal(self._pupil, other._pupil, equal_nan=True):
            return False
        if not np.array_equal(self._viewer_distance, other._viewer_distance, equal_nan=True):
            return False
        if not np.array_equal(self._pixel_size, other._pixel_size, equal_nan=True):
            return False
        return True

    def __str__(self) -> str:
        return f"{self.label.name}({self.duration}ms)"

    def __repr__(self) -> str:
        return self.__str__()

    def __len__(self) -> int:
        return self.num_samples


class FixationEvent(BaseEvent):
    _LABEL = EventLabelEnum.FIXATION


class SaccadeEvent(BaseEvent):
    _LABEL = EventLabelEnum.SACCADE


class PSOEvent(BaseEvent):
    _LABEL = EventLabelEnum.PSO


class SmoothPursuitEvent(BaseEvent):
    _LABEL = EventLabelEnum.SMOOTH_PURSUIT


class BlinkEvent(BaseEvent):
    _LABEL = EventLabelEnum.BLINK


EventSequenceType = Sequence[BaseEvent]



import peyes._utils.constants as constants

from peyes._utils.event_utils import parse_label
from peyes._base.parse import parse as parse_data
from peyes._base.set_config import set_viewer_distance, set_screen_monitor, set_event_configurations
from peyes._base.create import create_detector, create_events, create_boolean_channel
from peyes._base.match import match
from peyes._base.postprocess_events import summarize_events, events_to_labels

import peyes.datasets as datasets
import peyes.event_metrics as event_metrics
import peyes.sample_metrics as sample_metrics
import peyes.match_metrics as match_metrics
import peyes.alignment_metrics as channel_metrics
import peyes.visualize as visualization




from typing import Optional, Union

import numpy as np

from peyes._DataModels.EventLabelEnum import EventLabelEnum
from peyes._DataModels.Event import EventSequenceType
from peyes._DataModels.EventMatcher import OneToOneEventMatchesType
from peyes._DataModels.UnparsedEventLabel import UnparsedEventLabelType, UnparsedEventLabelSequenceType

from peyes._utils.event_utils import parse_label
from peyes._utils.metric_utils import dprime_and_criterion


def match_ratio(
        prediction: EventSequenceType,
        matches: OneToOneEventMatchesType,
        labels: Optional[Union[UnparsedEventLabelType, UnparsedEventLabelSequenceType]] = None,
) -> float:
    """
    Calculates the ratio of matched events to the total number of predicted events, optionally filtered by the given
    event labels. Returns NaN if there are no predicted events.

    :param prediction: all predicted events
    :param matches: the one-to-one matches between (subset of) ground-truth and (subset of) predicted events
    :param labels: optional event-labels to consider when calculating the ratio

    :return: the match ratio
    """
    if len(prediction) == 0:
        return np.nan
    if labels is None:
        labels = set(EventLabelEnum)
    elif isinstance(labels, UnparsedEventLabelType):
        labels = {parse_label(labels)}
    else:
        labels = set(parse_label(l) for l in labels)
    num_matched = sum(1 for e in matches.values() if e.label in labels)
    num_predicted = sum(1 for e in prediction if e.label in labels)
    return num_matched / num_predicted if num_predicted > 0 else np.nan


def precision_recall_f1(
        ground_truth: EventSequenceType,
        prediction: EventSequenceType,
        matches: OneToOneEventMatchesType,
        positive_label: Optional[Union[UnparsedEventLabelType, UnparsedEventLabelSequenceType]],
) -> (float, float, float):
    """
    Calculates the precision, recall, and F1-score for the given ground-truth and predicted events, where successfully
    matched events are considered true positives. The provided labels are considered as "positive" events.

    :param ground_truth: all ground-truth events
    :param prediction: all predicted events
    :param matches: the one-to-one matches between (subset of) ground-truth and (subset of) predicted events
    :param positive_label: event-label(s) to consider as "positive" events
    :return: the precision, recall, and F1-score values
    """
    p, n, pp, tp = _extract_contingency_values(ground_truth, prediction, matches, positive_label)
    recall = tp / p if p > 0 else np.nan
    precision = tp / pp if pp > 0 else np.nan
    f1 = 2 * (precision * recall) / (precision + recall) if precision + recall > 0 else np.nan
    return precision, recall, f1


def false_alarm_rate(
        ground_truth: EventSequenceType,
        prediction: EventSequenceType,
        matches: OneToOneEventMatchesType,
        positive_label: Optional[Union[UnparsedEventLabelType, UnparsedEventLabelSequenceType]],
) -> float:
    """
    Calculates the false-alarm rate for the given ground-truth and predicted events, where successfully matched events
    are considered true positives. The provided labels are considered as "positive" events.
    Note: False-Alarm rate could exceed 1.0 if there are many false alarms (pp - tp) and few negative GT events (n).
    # TODO: Consider adding a "correction" parameter to handle this case.

    :param ground_truth: all ground-truth events
    :param prediction: all predicted events
    :param matches: the one-to-one matches between (subset of) ground-truth and (subset of) predicted events
    :param positive_label: event-label(s) to consider as "positive" events
    :return: the false-alarm rate value
    """
    p, n, pp, tp = _extract_contingency_values(ground_truth, prediction, matches, positive_label)
    if n <= 0:
        return np.nan
    if 0 <= pp - tp <= n:
        return (pp - tp) / n
    # FA rate exceed 1.0 due to over-predicted false-alarm events
    return np.nan


def d_prime_and_criterion(
        ground_truth: EventSequenceType,
        prediction: EventSequenceType,
        matches: OneToOneEventMatchesType,
        positive_label: Optional[Union[UnparsedEventLabelType, UnparsedEventLabelSequenceType]],
        correction: Optional[str] = "loglinear",
) -> float:
    """
    Calculates d-prime and criterion while optionally applying a correction for floor/ceiling effects on the hit-rate
    and/or false-alarm rate. See information on correction methods at https://stats.stackexchange.com/a/134802/288290.
    See implementation details at https://lindeloev.net/calculating-d-in-python-and-php/.

    :param ground_truth: all ground-truth events
    :param prediction: all predicted events
    :param matches: the one-to-one matches between (subset of) ground-truth and (subset of) predicted events
    :param positive_label: event-label to consider as "positive" events
    :param correction: optional floor/ceiling correction method when calculating hit-rate and false-alarm rate
    :return:
        - d_prime: float; the d-prime value
        - criterion: float; the criterion value
    """
    p, n, pp, tp = _extract_contingency_values(ground_truth, prediction, matches, positive_label)
    return dprime_and_criterion(p, n, pp, tp, correction)


def _extract_contingency_values(
        ground_truth: EventSequenceType,
        prediction: EventSequenceType,
        matches: OneToOneEventMatchesType,
        positive_label: Optional[Union[UnparsedEventLabelType, UnparsedEventLabelSequenceType]],
) -> (int, int, int, int):
    """
    Extracts contingency values, used to fill in the confusion matrix for the provided matches between ground-truth and
    predicted events, where the given labels are considered "positive" events.

    :return:
        p: int; number of positive GT events
        n: int; number of negative GT events
        pp: int; number of positive predicted events
        tp: int; number of true positive predictions
    """
    if isinstance(positive_label, UnparsedEventLabelType):
        positive_label = {positive_label}
    positive_label = set(parse_label(l) for l in positive_label)
    if positive_label == set(EventLabelEnum):
        raise ValueError("Cannot consider all event labels as `positive` events.")
    p = len([e for e in ground_truth if e.label in positive_label])
    n = len(ground_truth) - p
    pp = len([e for e in prediction if e.label in positive_label])
    tp = len([e for e in matches.values() if e.label in positive_label])
    return p, n, pp, tp



from typing import Tuple, Union
from collections import Counter

import numpy as np
import plotly.graph_objects as go
from scipy.ndimage import gaussian_filter
from plotly.subplots import make_subplots

import peyes._utils.constants as cnst
import peyes._utils.visualization_utils as vis_utils
from peyes._utils.vector_utils import is_one_dimensional
from peyes._utils.pixel_utils import cast_to_integers

# TODO: create figure of y-coordinates over x-coordinates with color changing according to event labels


def gaze_heatmap(
        x: np.ndarray,
        y: np.ndarray,
        resolution: Tuple[int, int],
        **kwargs,
) -> go.Figure:
    """
    Creates a heatmap of the given x and y coordinates, overlayed on a background image or a white background.

    :param x: 1D array of x coordinates
    :param y: 1D array of y coordinates
    :param resolution: tuple of (width, height) in pixels

    :keyword bg_image: the background image to overlay on, defaults to None.
    :keyword bg_image_format: the color format of the background image (if provided), defaults to "BGR".
    :keyword bg_color: the background color if no image is provided, defaults to white.
    :keyword bg_alpha: the alpha value of the background image (range [0, 1]), defaults to 1.
    :keyword sigma: standard deviation of the Gaussian filter that smooths the heatmap, defaults to 10.0
    :keyword scale: scale factor for the pixel counts, defaults to sigma^2 (to match the Gaussian filter)
    :keyword colorscale: name of the color scale to use. Must be one of the named color scales in plotly.express.colors
    :keyword opacity: opacity of the heatmap (0-1). Default is 0.5

    :return: the figure.
    """
    x, y = __verify_same_length(x, y)
    bg = vis_utils.create_image(
        resolution,
        image=kwargs.get("bg_image", None),
        alpha=kwargs.get("bg_alpha", 1),
        color_format=kwargs.get("bg_image_format", "BGR"),
        default_color=kwargs.get("bg_color", "#ffffff"),
    )
    fig = go.Figure(
        data=go.Image(z=bg),
        layout=dict(width=resolution[0], height=resolution[1], margin=dict(l=0, r=0, b=0, t=0)),
    )
    sig = kwargs.get("sigma", 10.0)
    scale = kwargs.get("scale", sig**2)
    counts = scale * __pixel_counts(x, y, resolution)
    filtered_counts = gaussian_filter(counts, sigma=sig)
    heatmap = (filtered_counts - np.nanmin(filtered_counts)) / (np.nanmax(filtered_counts) - np.nanmin(filtered_counts))
    heatmap[(~np.isfinite(heatmap)) | (heatmap <= np.nanmedian(heatmap))] = np.nan    # remove low values
    fig.add_trace(go.Heatmap(
        z=heatmap,
        colorscale=kwargs.get("colorscale", "Jet"),
        opacity=kwargs.get("opacity", 0.5),
        showscale=False,
    ))
    fig.update_xaxes(
        visible=False, showticklabels=False, showgrid=False, showline=False, zeroline=False
    )
    fig.update_yaxes(
        visible=False, showticklabels=False, showgrid=False, showline=False, zeroline=False
    )
    fig.update_layout(
        paper_bgcolor='rgba(0, 0, 0, 0)', plot_bgcolor='rgba(0, 0, 0, 0)',  # transparent background
    )
    return fig


def gaze_over_time(
        x: np.ndarray,
        y: np.ndarray,
        t: np.ndarray,
        v: np.ndarray = None,
        vert_lines: np.ndarray = None,
        title: str = "Gaze Coordinates Over Time",
        **kwargs,
) -> go.Figure:
    """
    Plots the x- and y-coordinates of the gaze over time, optionally with velocity as a secondary y-axis and/or vertical
    lines at specified time points.

    :param x: gaze x-coordinates.
    :param y: gaze y-coordinates.
    :param t: the time axis.
    :param v: gaze velocity, optional. If provided, will be plotted on a secondary y-axis. Note the units of the
        velocity are px/s by default, see `v_measure`.
    :param vert_lines: time points to add vertical lines at, optional.
    :param title: the title of the figure.

    :keyword mode: 'lines'/'markers'/'lines+markers', defaults to 'lines'.
    :keyword line_width: width of the plotted lines, if `mode` contains 'lines'. defaults to 2.
    :keyword marker_size: the size of the markers, if `mode` contains 'markers'. defaults to 4.
    :keyword x_color: the color of the x-coordinates, defaults to red.
    :keyword y_color: the color of the y-coordinates, defaults to blue.
    :keyword v_color: the color of the velocity, defaults to light gray.
    :keyword v_measure: the measure of the velocity, defaults to "px/s".
    :keyword vert_line_colors: the colors of the vertical lines, should either be a single color or a list of colors the
        same length as `vert_lines`. Defaults to black.
    :keyword vert_line_width: the width of the vertical lines, defaults to 1.

    :return: the figure.
    """
    x, y, t, v = __verify_same_length(x, y, t, v)
    mode = kwargs.get('mode', 'lines')
    line_width = kwargs.get('line_width', 2)
    marker_size = kwargs.get("marker_size", 4)
    fig = make_subplots(specs=[[{"secondary_y": True}]])
    fig.add_trace(
        go.Scatter(
            x=t, y=x, name="x",
            mode=mode,
            line=dict(color=kwargs.get("x_color", "#ff0000"), width=line_width),
            marker=dict(color=kwargs.get("x_color", "#ff0000"), size=marker_size),
        ), secondary_y=False,
    )
    fig.add_trace(
        go.Scatter(
            x=t, y=y, name="y",
            mode=mode,
            line=dict(color=kwargs.get("y_color", "#0000ff"), width=line_width),
            marker=dict(color=kwargs.get("y_color", "#0000ff"), size=marker_size),
        ), secondary_y=False,
    )

    if v is not None:
        v_measure = kwargs.get("v_measure", "px/s")
        y_axis2_title = f"{cnst.VELOCITY_STR} ({v_measure})"
        fig.add_trace(
            go.Scatter(
                x=t, y=v, name="v",
                mode=mode,
                line=dict(color=kwargs.get("v_color", "#bbbbbb"), width=line_width),
                marker=dict(color=kwargs.get("v_color", "#bbbbbb"), size=marker_size),
            ), secondary_y=True,
        )

        # align y-axis scales on (0_1, 0_2) and (max-val_1, max-val_2)
        v_max = np.nanmax(v)
        xy_max = max(np.nanmax(x), np.nanmax(y))
        fig.update_layout(yaxis2=dict(scaleanchor='y', scaleratio=xy_max/v_max, constraintoward='bottom'))

    else:
        y_axis2_title = None

    if vert_lines is not None:
        vert_line_color = kwargs.get("vert_line_color", ["#000000"] * len(vert_lines))
        vert_line_color = [vert_line_color] * len(vert_lines) if isinstance(vert_line_color, str) else vert_line_color
        assert len(vert_lines) == len(vert_line_color), "Length mismatch: `vert_lines` and `vert_line_color`"
        vert_line_width = kwargs.get("vert_line_width", 1)
        for v, c in zip(vert_lines, vert_line_color):
            fig.add_vline(x=v, line=dict(color=c, width=vert_line_width, dash="dash"))

    fig.update_layout(
        title=title,
        xaxis_title=f"{cnst.TIME_STR} ({cnst.SAMPLE_STR})",
        yaxis_title=f"{cnst.PIXEL_STR} coordinates",
        yaxis2_title=y_axis2_title,
    )
    return fig


def gaze_trajectory(
        x: np.ndarray,
        y: np.ndarray,
        t: np.ndarray,
        resolution: Tuple[int, int],
        title: str = "Gaze Visualization",
        **kwargs,
) -> go.Figure:
    return _visualize_gaze_trajectory(
        x=x,
        y=y,
        resolution=resolution,
        title=title,
        bg_image=kwargs.get("bg_image", None),
        bg_image_format=kwargs.get("bg_image_format", "BGR"),
        bg_alpha=kwargs.get("bg_alpha", 1),
        bg_color=kwargs.get("bg_color", "#ffffff"),
        marker_color=t,
        marker_size=kwargs.get("marker_size", 5),
        marker_alpha=kwargs.get("marker_alpha", 1),
        marker_colorscale=kwargs.get("colorscale", "Jet"),
        marker_colorbar_title=cnst.TIME_STR.title(),
    )



def _visualize_gaze_trajectory(
        x: np.ndarray,
        y: np.ndarray,
        resolution: Tuple[int, int],
        title: str = None,
        bg_image: np.ndarray = None,
        bg_image_format: str = "BGR",
        bg_alpha: float = 1,
        bg_color: str = "#ffffff",
        marker_color: Union[str, np.ndarray] = "#000000",
        marker_size: Union[int, np.ndarray] = 5,
        marker_alpha: Union[float, np.ndarray] = 1,
        marker_colorscale: str = "Jet",
        marker_colorbar_title: str = None,
) -> go.Figure:
    if isinstance(marker_color, str):
        marker_color = np.full_like(x, marker_color)
    if isinstance(marker_size, int) or isinstance(marker_size, float):
        marker_size = np.full_like(x, marker_size)
    if isinstance(marker_alpha, int) or isinstance(marker_alpha, float):
        marker_alpha = np.full_like(x, marker_alpha)
    x, y, marker_color, marker_size, marker_alpha = __verify_markers(x, y, marker_color, marker_size, marker_alpha)
    bg = vis_utils.create_image(
        resolution, image=bg_image, color_format=bg_image_format, alpha=bg_alpha, default_color=bg_color
    )
    fig = go.Figure(
        data=go.Image(z=bg, colormodel='rgba256'),
        layout=dict(width=resolution[0], height=resolution[1], margin=dict(l=0, r=0, b=0, t=0)),
    )
    fig.add_trace(go.Scatter(
        x=x, y=y, mode="markers", marker=dict(
            color=marker_color,
            size=marker_size,
            opacity=marker_alpha,
            colorscale=marker_colorscale,
            colorbar=dict(title=marker_colorbar_title),
            line=dict(width=0),
        )
    ))
    fig.update_layout(
        title=title, title_y=0.98, title_yanchor='top',
        xaxis=dict(visible=False, showticklabels=False, showgrid=False, showline=False, zeroline=False),
        yaxis=dict(visible=False, showticklabels=False, showgrid=False, showline=False, zeroline=False),
        paper_bgcolor='rgba(0, 0, 0, 0)', plot_bgcolor='rgba(0, 0, 0, 0)',  # transparent background
    )
    return fig


def __verify_markers(
        x: np.ndarray, y: np.ndarray, color: np.ndarray, size: np.ndarray, alpha: np.ndarray,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """ Verifies all marker arrays have valid values and the same length as the x and y arrays. """
    if not all(size >= 0):
        raise ValueError("`size` must be non-negative")
    if not all((0 <= alpha) & (alpha <= 1)):
        raise ValueError("`alpha` must be in the range [0, 1]")
    x, y, color, size, alpha = __verify_same_length(x, y, color, size, alpha)
    return x, y, color, size, alpha


def __verify_same_length(*arrays: np.ndarray) -> Tuple[np.ndarray, ...]:
    """  Verifies all input arrays are one dimensional and have the same length.  """
    if not all(is_one_dimensional(arr) for arr in arrays):
        raise ValueError("All arrays must be one-dimensional")
    if not all(len(arr) == len(arrays[0]) for arr in arrays):
        raise ValueError("All arrays must have the same length")
    return tuple(arr.reshape(-1) for arr in arrays)


def __pixel_counts(
        x: np.ndarray,
        y: np.ndarray,
        resolution: Tuple[int, int],
) -> np.ndarray:
    w, h = resolution
    counts = np.zeros((h, w), dtype=int)
    x_int, y_int = cast_to_integers(x, y)
    counter = Counter(zip(y_int, x_int))
    for (y_, x_), count in counter.items():
        if not np.isfinite(x_) or not np.isfinite(y_):
            continue
        counts[int(y_), int(x_)] = count
    return counts



import os
from typing import List, Optional, Union, Sequence, Callable, Any

import numpy as np
import pandas as pd
from tqdm import tqdm

import peyes
from peyes._DataModels.UnparsedEventLabel import UnparsedEventLabelType, UnparsedEventLabelSequenceType
from peyes._DataModels.EventLabelEnum import EventLabelEnum
from peyes._DataModels.Event import EventSequenceType
from peyes._utils.event_utils import calculate_sampling_rate

import analysis.utils as u
import analysis.process._helpers as h

_CHANNEL_TYPE_STR = "channel_type"


def run_default(
        dataset_name: str,
        dataset: pd.DataFrame,
        pos_labels: Optional[Union[UnparsedEventLabelType, UnparsedEventLabelSequenceType]] = None,
):
    default_output_dir = h.get_default_output_dir(dataset_name)
    try:
        events = pd.read_pickle(os.path.join(default_output_dir, f"{peyes.constants.EVENTS_STR}.pkl"))
    except FileNotFoundError:
        raise FileNotFoundError(
            f"Couldn't find `{peyes.constants.EVENTS_STR}.pkl` in {default_output_dir}. Please preprocess the dataset first."
        )
    channel_metrics_dir = os.path.join(default_output_dir, peyes.constants.EVENTS_CHANNEL_STR)
    os.makedirs(channel_metrics_dir, exist_ok=True)
    time_diffs_fullpath = os.path.join(
        channel_metrics_dir, u.get_filename_for_labels(pos_labels, suffix="timing_differences", extension="pkl")
    )
    try:
        time_diffs = pd.read_pickle(time_diffs_fullpath)
    except FileNotFoundError:
        time_diffs = timing_differences(dataset, events, u.DATASET_ANNOTATORS[dataset_name], pos_labels=pos_labels)
        time_diffs.to_pickle(time_diffs_fullpath)
    sdt_metrics_fullpath = os.path.join(
        channel_metrics_dir, u.get_filename_for_labels(pos_labels, suffix="sdt_metrics", extension="pkl")
    )
    try:
        sdt_metrics = pd.read_pickle(sdt_metrics_fullpath)
    except FileNotFoundError:
        sdt_metrics = detection_metrics(dataset, events, np.arange(10), u.DATASET_ANNOTATORS[dataset_name], pos_labels=pos_labels)
        sdt_metrics.to_pickle(sdt_metrics_fullpath)
    return time_diffs, sdt_metrics


def timing_differences(
        dataset: pd.DataFrame,
        events: pd.DataFrame,
        gt_labelers: List[str],
        pred_labelers: List[str] = None,
        pos_labels: Optional[Union[UnparsedEventLabelType, UnparsedEventLabelSequenceType]] = None,
) -> pd.DataFrame:
    results = _calculation_wrapper(
        dataset,
        events,
        peyes.channel_metrics.onset_differences,
        peyes.channel_metrics.offset_differences,
        gt_labelers,
        pred_labelers,
        pos_labels,
        iteration_desc="Events Channel :: Timing Differences",
    )
    results = pd.DataFrame.from_dict(results, orient="columns")
    results.columns.names = [peyes.constants.TRIAL_ID_STR, u.GT_STR, u.PRED_STR, peyes.constants.ITERATION_STR]
    results.index.names = [_CHANNEL_TYPE_STR]
    return results


def detection_metrics(
        dataset: pd.DataFrame,
        events: pd.DataFrame,
        threshold: Union[int, Sequence[int]],
        gt_labelers: List[str],
        pred_labelers: List[str] = None,
        pos_labels: Optional[Union[UnparsedEventLabelType, UnparsedEventLabelSequenceType]] = None,
        dprime_correction: str = "loglinear",
) -> pd.DataFrame:
    onset_func = lambda gt, pred, sr, ns: peyes.channel_metrics.onset_detection_metrics(
        gt, pred, threshold, sr, ns, dprime_correction
    )
    offset_func = lambda gt, pred, sr, ns: peyes.channel_metrics.offset_detection_metrics(
        gt, pred, threshold, sr, ns, dprime_correction
    )
    results = _calculation_wrapper(
        dataset,
        events,
        onset_func,
        offset_func,
        gt_labelers,
        pred_labelers,
        pos_labels,
        iteration_desc="Events Channel :: SDT Metrics",
    )
    results = pd.concat({k: pd.concat(v, axis=0) for k, v in results.items()}, axis=1)
    results.columns.names = [
        peyes.constants.TRIAL_ID_STR, u.GT_STR, u.PRED_STR, peyes.constants.ITERATION_STR, peyes.constants.METRIC_STR
    ]
    results.index.names = [_CHANNEL_TYPE_STR, peyes.constants.THRESHOLD_STR]
    return results


def _calculation_wrapper(
        dataset: pd.DataFrame,
        events: pd.DataFrame,
        onset_func: Callable[[EventSequenceType, EventSequenceType, float, int], Any],
        offset_func: Callable[[EventSequenceType, EventSequenceType, float, int], Any],
        gt_labelers: List[str],
        pred_labelers: List[str] = None,
        pos_labels: Optional[Union[UnparsedEventLabelType, UnparsedEventLabelSequenceType]] = None,
        iteration_desc: str = "Channel Metrics",
) -> dict:
    pred_labelers = h.check_labelers(events, pred_labelers)
    if pos_labels is None:
        pos_labels = [l for l in EventLabelEnum]
    elif isinstance(pos_labels, UnparsedEventLabelType):
        pos_labels = [peyes.parse_label(pos_labels)]
    else:
        pos_labels = [peyes.parse_label(l) for l in pos_labels]

    results = dict()
    trials = events.columns.get_level_values(level=peyes.constants.TRIAL_ID_STR).unique()
    for tr in tqdm(trials, desc=iteration_desc):
        trial_timestamps = dataset[dataset[peyes.constants.TRIAL_ID_STR] == tr][peyes.constants.T].values
        trial_num_samples = len(trial_timestamps)
        trial_sampling_rate = calculate_sampling_rate(trial_timestamps)
        for gt_lblr in gt_labelers:
            try:
                trial_gt_events = events.xs((tr, gt_lblr), axis=1, level=[peyes.constants.TRIAL_ID_STR, peyes.constants.LABELER_STR])
            except KeyError:
                continue
            if trial_gt_events.size == 0:
                continue
            gt_min_iteration = np.nanmin(trial_gt_events.columns.get_level_values(peyes.constants.ITERATION_STR))
            gt_events = events[tr, gt_lblr, gt_min_iteration].dropna().values.flatten()
            gt_events = np.array([e for e in gt_events if e.label in pos_labels])
            if gt_events.size == 0:
                continue
            for pred_lblr in pred_labelers:
                try:
                    pred_events_all_iters = events.xs(
                        (tr, pred_lblr), axis=1, level=[peyes.constants.TRIAL_ID_STR, peyes.constants.LABELER_STR]
                    )
                except KeyError:
                    continue
                for pred_it in pred_events_all_iters.columns.get_level_values(peyes.constants.ITERATION_STR).unique():
                    if (pred_lblr == gt_lblr) and (pred_it == gt_min_iteration):
                        continue
                    pred_events = pred_events_all_iters[pred_it].dropna().values.flatten()
                    pred_events = np.array([e for e in pred_events if e.label in pos_labels])
                    if pred_events.size == 0:
                        continue
                    results[(tr, gt_lblr, pred_lblr, pred_it)] = {
                        "onset": onset_func(gt_events, pred_events, trial_sampling_rate, trial_num_samples),
                        "offset": offset_func(gt_events, pred_events, trial_sampling_rate, trial_num_samples)
                    }
    return results



from typing import Optional

import numpy as np

import peyes._utils.constants as cnst
from peyes._DataModels.EventLabelEnum import EventLabelEnum
from peyes._DataModels.Event import BaseEvent, EventSequenceType
from peyes._DataModels.UnparsedEventLabel import UnparsedEventLabelType


def calculate_sampling_rate(ms: np.ndarray) -> float:
    """
    Calculates the sampling rate of the given timestamps in Hz.
    :param ms: timestamps in milliseconds (floating-point, not integer)
    """
    if len(ms) < 2:
        raise ValueError("timestamps must be of length at least 2")
    sr = cnst.MILLISECONDS_PER_SECOND / np.mean(np.diff(ms))
    if not np.isfinite(sr):
        raise RuntimeError("Error calculating sampling rate")
    return round(sr, 1)


def calculate_num_samples(
        start_time: float, end_time: float, sampling_rate: float, min_samples: int = 1
) -> int:
    """
    Calculates the number of samples between the given start and end times, given the sampling rate. The number of
    samples is rounded up to the nearest integer, but at least `min_samples` are returned.

    :param start_time: the start time in milliseconds
    :param end_time: the end time in milliseconds
    :param sampling_rate: the sampling rate in Hz
    :param min_samples: the minimum number of samples to return

    :return: the number of samples

    :raises ValueError: if the start or end times are invalid (not finite or start > end)
    :raises ValueError: if the sampling rate is invalid (not finite or non-positive)
    :raises RuntimeError: if an error occurs during the calculation
    """
    if not np.isfinite(start_time) or not np.isfinite(end_time) or start_time > end_time:
        raise ValueError("Invalid start or end time")
    if not np.isfinite(sampling_rate) or sampling_rate <= 0:
        raise ValueError("sampling rate must be a positive finite number")
    res = np.ceil((end_time - start_time) * sampling_rate / cnst.MILLISECONDS_PER_SECOND)
    if not np.isfinite(res) or res < 0:
        raise RuntimeError("Error calculating number of samples")
    return max(int(res), min_samples)


def parse_label(
        val: Optional[UnparsedEventLabelType],
        safe: bool = True
) -> EventLabelEnum:
    """
    Parses a gaze label from the original dataset's type to an EventLabelEnum
    :param val: the value to parse
    :param safe: if True, returns EventLabelEnum.UNDEFINED when the parsing fails; otherwise, raises an exception
    :return: the parsed gaze label
    """
    try:
        if isinstance(val, EventLabelEnum):
            return val
        if isinstance(val, BaseEvent):
            return val.label
        if isinstance(val, int):
            return EventLabelEnum(val)
        if isinstance(val, str):
            if val.isdigit():
                return EventLabelEnum(int(val))
            if "." in val or "," in val:
                # Handle cases like "1.0", "2,0", etc.
                return EventLabelEnum(float(val))
            if " " in val:
                val = val.replace(" ", "_")
            return EventLabelEnum[val.upper()]
        if isinstance(val, float):
            if not val.is_integer():
                raise ValueError(f"Invalid value: {val}")
            return EventLabelEnum(int(val))
        if np.isscalar(val) and isinstance(val, np.number):
            return parse_label(float(val), safe)
        raise TypeError(f"Incompatible type: {type(val)}")
    except Exception as err:
        if safe and (isinstance(err, ValueError) or isinstance(err, KeyError) or isinstance(err, TypeError)):
            return EventLabelEnum.UNDEFINED
        raise err


def microsaccade_ratio(events: EventSequenceType, amplitude_threshold) -> float:
    """
    Calculates the ratio of microsaccades to saccades. Returns NaN if there are no saccades.
    :param events: sequence of events
    :param amplitude_threshold: threshold for microsaccades (degrees)
    :return: the ratio of microsaccades to saccades
    """
    saccades = [e for e in events if e.label == EventLabelEnum.SACCADE]
    if len(saccades) == 0:
        return np.nan
    microsaccades = [e for e in saccades if e.amplitude < amplitude_threshold]
    return len(microsaccades) / len(saccades)



import os
from typing import List

import pandas as pd

import peyes

import analysis.utils as u

_DEFAULT_STR = "default"


def get_default_output_dir(dataset_name: str) -> str:
    res = os.path.join(u.OUTPUT_DIR, dataset_name, _DEFAULT_STR)
    os.makedirs(res, exist_ok=True)
    return res


def check_labelers(data: pd.DataFrame, labelers: List[str] = None) -> List[str]:
    available_labelers = data.columns.get_level_values(peyes.constants.LABELER_STR).unique()
    labelers = set(labelers or available_labelers)
    unknown_labelers = labelers - set(available_labelers)
    if unknown_labelers:
        raise ValueError(f"Unknown labelers: {unknown_labelers}")
    return list(labelers)



from typing import Union

import numpy as np
import pandas as pd

import peyes._utils.constants as cnst


def parse(
        data: Union[np.ndarray, pd.DataFrame],
        time_name: str = cnst.T,
        x_name: str = cnst.X,
        y_name: str = cnst.Y,
        pupil_name: str = cnst.PUPIL,
        missing_data_value: float = np.nan,
):
    """
    Parse raw gaze data into a DataFrame, and rename the columns to the standard names: `t`, `x`, `y`, `pupil`.
    Missing columns are filled with NaN, and missing data is replaced with NaN.

    :param data: DataFrame or named-array containing the raw gaze data
    :param time_name: name of the time column (ms)
    :param x_name: name of the horizontal gaze position column (pixels)
    :param y_name: name of the vertical gaze position column (pixels)
    :param pupil_name: name of the pupil size column (mm)
    :param missing_data_value: the value used to represent missing data in the raw data

    :return: a DataFrame containing the parsed gaze data with renamed to columns to the standard names
    """
    if isinstance(data, np.ndarray):
        data = pd.DataFrame(data)
    elif isinstance(data, pd.DataFrame):
        data = data.copy(deep=True)  # avoid overwriting the original dataset
    else:
        raise TypeError(f"Invalid dataset type: {type(data)}")

    # rename columns
    columns_rename = {time_name: cnst.T, x_name: cnst.X, y_name: cnst.Y, pupil_name: cnst.PUPIL}
    data.rename(columns=columns_rename, inplace=True)
    missing_columns = [col for col in columns_rename.values() if col not in data.columns]
    for col in missing_columns:
        data.loc[:, col] = np.nan

    # replace missing data with NaN
    if np.isnan(missing_data_value):
        nan_idxs = data[(data[cnst.X].isna()) | (data[cnst.Y].isna())].index
    else:
        nan_idxs = data[(data[cnst.X] == missing_data_value) | (data[cnst.Y] == missing_data_value)].index
    no_status_idxs = data[~data[cnst.STATUS_STR]].index if cnst.STATUS_STR in data.columns else []
    nan_idxs = np.union1d(nan_idxs, no_status_idxs)
    data.loc[nan_idxs, [cnst.X, cnst.Y]] = np.nan
    return data



from typing import List, Optional, Union, Tuple, Sequence

import pandas as pd
import plotly.graph_objects as go

import peyes
from peyes._DataModels.UnparsedEventLabel import UnparsedEventLabelType, UnparsedEventLabelSequenceType

import analysis.utils as u
import analysis.statistics._helpers as h

###################


def load_sdt(
        dataset_name: str,
        output_dir: str,
        label: Optional[Union[UnparsedEventLabelType, UnparsedEventLabelSequenceType]] = None,
        stimulus_type: Optional[Union[str, List[str]]] = None,
        metric: Optional[Union[str, List[str]]] = None,
        iteration: int = 1,
) -> pd.DataFrame:
    return h.load_data(
        dataset_name=dataset_name, output_dir=output_dir,
        data_dir_name=f"{peyes.constants.SAMPLE_STR}_{peyes.constants.METRICS_STR}",
        filename_suffix=f"{u.SDT_STR}_{peyes.constants.METRICS_STR}", label=label,
        iteration=iteration, stimulus_type=stimulus_type, sub_index=metric,
    )


def load_global_metrics(
        dataset_name: str,
        output_dir: str,
        stimulus_type: Optional[Union[str, List[str]]] = None,
        metric: Optional[Union[str, List[str]]] = None,
        iteration: int = 1,
) -> pd.DataFrame:
    return h.load_data(
        dataset_name=dataset_name, output_dir=output_dir,
        data_dir_name=f"{peyes.constants.SAMPLE_STR}_{peyes.constants.METRICS_STR}",
        filename_suffix=f"{u.GLOBAL_STR}_{peyes.constants.METRICS_STR}", label=None,
        iteration=iteration, stimulus_type=stimulus_type, sub_index=metric,
    )


def wilcoxon(
        data: pd.DataFrame,
        gt_cols: Union[str, Sequence[str]],
        alternative: str = "two-sided",
        method: str = "auto",
) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    return h.wilcoxon(data, gt_cols, alternative, method)


def kruskal_wallis_dunns(
        data: pd.DataFrame,
        gt_cols: Union[str, Sequence[str]],
        multi_comp: Optional[str] = "fdr_bh",
) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    return h.kruskal_wallis_dunns(data, gt_cols, multi_comp)


def friedman_nemenyi(
        data: pd.DataFrame,
        gt_cols: Union[str, Sequence[str]],
) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    return h.friedman_nemenyi(data, gt_cols)


def post_hoc_table(
        ph_data: pd.DataFrame,
        metric: str,
        gt_cols: Union[str, Sequence[str]],
        alpha: float = 0.05,
        marginal_alpha: Optional[float] = 0.075,
) -> pd.DataFrame:
    if isinstance(gt_cols, str):
        gt_cols = [gt_cols]
    return h.create_post_hoc_table(ph_data, metric, *gt_cols, alpha=alpha, marginal_alpha=marginal_alpha)


def sdt_distributions_figure(
        data: pd.DataFrame,
        gt1: str,
        gt2: str,
        colors: u.COLORMAP_TYPE = None,
        title: str = "",
        only_box: bool = False,
        show_other_gt: bool = False,
        share_x: bool = False,
        share_y: bool = False,
) -> go.Figure:
    title = title or "Samples :: SDT Metrics Distributions"
    return h.distributions_figure(
        data, gt1=gt1, gt2=gt2, colors=colors, title=title, only_box=only_box,
        show_other_gt=show_other_gt, share_x=share_x, share_y=share_y
    )


def global_metrics_distributions_figure(
        data: pd.DataFrame,
        gt1: str,
        gt2: str,
        colors: u.COLORMAP_TYPE = None,
        title: str = "",
        only_box: bool = False,
        show_other_gt: bool = False,
        share_x: bool = False,
        share_y: bool = False,
) -> go.Figure:
    title = title or f"Samples :: Global Metrics Distributions <br><sup>({peyes.constants.LABEL_STR.title()}:All)</sup>"
    return h.distributions_figure(
        data, gt1=gt1, gt2=gt2, colors=colors, title=title, only_box=only_box,
        show_other_gt=show_other_gt, share_x=share_x, share_y=share_y
    )



import pandas as pd
import plotly.io as pio

import peyes
import analysis.statistics.sample_metrics as sm
from analysis._default_values._helpers import *
from peyes._DataModels.EventLabelEnum import EventLabelEnum

pio.renderers.default = "browser"

# GLOBAL METRICS
######################

# Load Data
sample_global_metrics = sm.load_global_metrics(
    DATASET_NAME, PROCESSED_DATA_DIR, stimulus_type=STIMULUS_TYPE, metric=None
)
sample_global_metrics.drop(index=peyes.constants.ACCURACY_STR, inplace=True)    # Drop Accuracy metric

# Statistics: Kruskal-Wallis and Dunn's Post-Hoc
sm_global_statistics, sm_global_pvalues, sm_global_dunns, sm_global_Ns = sm.kruskal_wallis_dunns(
    sample_global_metrics, [GT1, GT2], multi_comp=MULTI_COMP
)

# Show Figure
sm_global_metrics_fig = sm.global_metrics_distributions_figure(sample_global_metrics, GT1, gt2=GT2, only_box=False)
sm_global_metrics_fig.update_layout(width=1600, height=550,)
sm_global_metrics_fig.show()

# Save Figure
peyes.visualize.save_figure(
    sm_global_metrics_fig, "sample_metrics-global", FIGURES_DIR, as_png=True, as_html=False, as_json=False
)

# SDT METRICS
######################

# Load Data for fixations & saccades
fixation_sdt = sm.load_sdt(
    DATASET_NAME, PROCESSED_DATA_DIR, label=EventLabelEnum.FIXATION.value, stimulus_type=STIMULUS_TYPE, metric=None
)
fixation_sdt = fixation_sdt.loc[[peyes.constants.D_PRIME_STR, peyes.constants.F1_STR]]  # Keep only d' and f1 metrics
fixation_sdt = fixation_sdt.rename(index=lambda idx: f"fixation_{idx}")     # Rename index

saccade_sdt = sm.load_sdt(
    DATASET_NAME, PROCESSED_DATA_DIR, label=EventLabelEnum.SACCADE.value, stimulus_type=STIMULUS_TYPE, metric=None
)
saccade_sdt = saccade_sdt.loc[[peyes.constants.D_PRIME_STR, peyes.constants.F1_STR]]  # Keep only d' and f1 metrics
saccade_sdt = saccade_sdt.rename(index=lambda idx: f"saccade_{idx}")     # Rename index

fixation_saccade_sdt = pd.concat([fixation_sdt, saccade_sdt], axis=0)
del fixation_sdt, saccade_sdt


# Statistics: Kruskal-Wallis and Dunn's Post-Hoc
sdt_statistics, sdt_pvalues, sdt_dunns, sdt_Ns = sm.kruskal_wallis_dunns(
    fixation_saccade_sdt, [GT1, GT2], multi_comp=MULTI_COMP
)

# Show Figures
sdt_metrics_fig = sm.sdt_distributions_figure(
    fixation_saccade_sdt, GT1, GT2, title="SDT Metrics (fixations & saccades)", only_box=False,
)
sdt_metrics_fig.update_layout(width=1600, height=550,)
sdt_metrics_fig.show()

# Save Figure
peyes.visualize.save_figure(
    sdt_metrics_fig, "sample_metrics-sdt-fixations&saccades", FIGURES_DIR, as_png=True, as_html=False, as_json=False
)



import os
import time
import copy
import warnings
from typing import List, Dict, Union

import numpy as np
import pandas as pd
from tqdm import tqdm, trange

import peyes
from peyes._DataModels.Detector import BaseDetector
from peyes._DataModels.UnparsedEventLabel import UnparsedEventLabelType, UnparsedEventLabelSequenceType

import analysis.utils as u
import analysis.process._helpers as h

_MATCH_BY_STR = "match_by"

DEFAULT_MATCHING_SCHEMES = {
    'iou': {_MATCH_BY_STR: 'iou', 'min_iou': 1/3},
    'max_overlap': {_MATCH_BY_STR: 'max_overlap', 'min_overlap': 0.5},
    'onset': {_MATCH_BY_STR: 'onset', 'max_onset_difference': 15},
    'offset': {_MATCH_BY_STR: 'offset', 'max_offset_difference': 15},
    'l2': {_MATCH_BY_STR: 'l2', 'max_l2': 15},
    # 'window': {_MATCH_BY_STR: 'window', 'max_onset_difference': 15, 'max_offset_difference': 15},
    **{f"window_{w}": {_MATCH_BY_STR: 'window', 'max_onset_difference': w, 'max_offset_difference': w} for w in np.arange(21)},
    # TODO: consider re-writing this to have `threshold` another argument
    # **{f"onset_{o}": {_MATCH_BY_STR: 'onset', 'max_onset_difference': o} for o in np.arange(21)},
    # **{f"iou_{iou:.1f}": {_MATCH_BY_STR: 'iou', 'min_iou': iou} for iou in np.arange(0.1, 1.01, 0.1)},
    # **{f"overlap_{ov:.1f}": {_MATCH_BY_STR: 'max_overlap', 'min_overlap': ov} for ov in np.arange(0.1, 1.01, 0.1)},
}

# peyes.match(gt_events, pred_events, match_by=match_by, allow_xmatch=allow_xmatch, **match_params)


def run_default(
        dataset_name: str, detectors: List[BaseDetector], verbose: bool = True
) -> (pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame):
    start = time.time()
    if verbose:
        print(f"Running full pipeline on {dataset_name}...")
    dataset = u.load_dataset(dataset_name, verbose=True)
    default_output_dir = h.get_default_output_dir(dataset_name)
    annotators = u.DATASET_ANNOTATORS[dataset_name]

    if verbose:
        print(f"Detecting labels & events...")
    labels, metadata, events = detect_labels_and_events(
        dataset, detectors, annotators, verbose=verbose
    )
    if verbose:
        print(f"Saving labels & events to {default_output_dir}...")
    labels.to_pickle(os.path.join(default_output_dir, f"{peyes.constants.LABELS_STR}.pkl"))
    metadata.to_pickle(os.path.join(default_output_dir, f"{peyes.constants.METADATA_STR}.pkl"))
    events.to_pickle(os.path.join(default_output_dir, f"{peyes.constants.EVENTS_STR}.pkl"))

    if verbose:
        print(f"Matching events...")
    matches = match_events(
        events, annotators, matching_schemes=None, allow_xmatch=False
    )
    if verbose:
        print(f"Saving matches to {default_output_dir}...")
    matches.to_pickle(os.path.join(default_output_dir, f"{peyes.constants.MATCHES_STR}.pkl"))
    elapsed = time.time() - start
    if verbose:
        print(f"### PREPROCESS TIME:\t{elapsed:.2f} seconds ###")
    return dataset, labels, metadata, events, matches


def detect_labels_and_events(
        dataset: pd.DataFrame,
        detectors: List[BaseDetector],
        annotators: List[str],
        num_iterations: int = 4,
        overwrite_label: Union[UnparsedEventLabelType, UnparsedEventLabelSequenceType] = 2,
        verbose: bool = True,
) -> (pd.DataFrame, pd.DataFrame, pd.DataFrame):
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        if overwrite_label is None and num_iterations > 1:
            raise ValueError("Cannot have multiple iterations without specifying a label to overwrite.")
        elif isinstance(overwrite_label, UnparsedEventLabelType):
            overwrite_label = {peyes.parse_label(overwrite_label)}
        else:
            overwrite_label = {peyes.parse_label(l) for l in overwrite_label}

        labels, metadata, events = dict(), dict(), dict()
        trials_prog_bar = tqdm(
            dataset[peyes.constants.TRIAL_ID_STR].unique(), desc="Detecting Labels & Events", leave=True, disable=False
        )
        for tr in trials_prog_bar:
            trial = dataset[dataset[peyes.constants.TRIAL_ID_STR] == tr]
            t = trial[peyes.constants.T].values
            x = trial[peyes.constants.X].values
            y = trial[peyes.constants.Y].values
            p = trial[peyes.constants.PUPIL].values
            vd = trial[peyes.constants.VIEWER_DISTANCE_STR].iloc[0]
            ps = trial[peyes.constants.PIXEL_SIZE_STR].iloc[0]
            for annot in tqdm(annotators, desc="\tAnnotators", leave=False, disable=not verbose, position=1):
                annotator_labels = trial[annot].values
                labels[(tr, annot, 1)] = np.array(annotator_labels)
                events[(tr, annot, 1)] = peyes.create_events(
                    labels=annotator_labels, t=t, x=x, y=y, pupil=p, viewer_distance=vd, pixel_size=ps
                )
            for det in tqdm(detectors, desc="\tDetectors", leave=False, disable=not verbose, position=1):
                det_name = det.name
                x_copy, y_copy, p_copy = copy.deepcopy(x), copy.deepcopy(y), copy.deepcopy(p)
                for it in trange(num_iterations, desc="\t\tIterations", leave=False, disable=not verbose, position=2):
                    it_labels, it_metadata = det.detect(t, x_copy, y_copy, viewer_distance_cm=vd, pixel_size_cm=ps)
                    labels[(tr, det_name, it+1)] = np.array(it_labels)
                    metadata[(tr, det_name, it+1)] = it_metadata
                    events[(tr, det_name, it+1)] = peyes.create_events(
                        labels=it_labels, t=t, x=x_copy, y=y_copy, pupil=p_copy, viewer_distance=vd, pixel_size=ps
                    )
                    to_overwrite = np.isin(it_labels, list(overwrite_label))
                    x_copy[to_overwrite] = np.nan
                    y_copy[to_overwrite] = np.nan
                    p_copy[to_overwrite] = np.nan
        labels = pd.concat([pd.Series(v, name=k) for k, v in labels.items()], axis=1)
        labels.index.name = peyes.constants.SAMPLE_STR
        metadata = pd.concat([pd.Series(v, name=k) for k, v in metadata.items()], axis=1)
        metadata.index.name = peyes.constants.FIELD_NAME_STR
        events = pd.concat([pd.Series(v, name=k) for k, v in events.items()], axis=1)
        events.index.name = peyes.constants.EVENT_STR
        labels.columns.names = metadata.columns.names = events.columns.names = [
            peyes.constants.TRIAL_ID_STR, peyes.constants.LABELER_STR, peyes.constants.ITERATION_STR
        ]
        return labels, metadata, events


def match_events(
        events: pd.DataFrame,
        gt_labelers: List[str],
        pred_labelers: List[str] = None,
        matching_schemes: Dict[str, Dict[str, Union[str, int, float]]] = None,
        allow_xmatch: bool = False,
):
    """
    Matches between ground-truth and predicted events for each trial, labeler, and iteration.

    :param events: pd.DataFrame with MultiIndex columns (trial, labeler, iteration)
    :param gt_labelers: labelers to use as ground-truth
    :param pred_labelers: labelers to use as predictors
    :param matching_schemes: dictionary matching different scheme names to scheme parameters. Each set of parameters
        must include the key `match_by` which specifies the scheme to match by. If not provided, the scheme name is
        used as the `match_by` value.
    :param allow_xmatch: if True, allows cross-matching, i.e. matching between GT and Pred events of different types.

    :return: pd.DataFrame where each row is a matching-scheme, and columns are MultiIndex (trial, gt_labeler,
        pred_labeler, iteration). Cells contain dictionaries matching (a subset of) GT events to (a subset of) Pred events.
    """
    pred_labelers = h.check_labelers(events, pred_labelers)
    matching_schemes = matching_schemes or DEFAULT_MATCHING_SCHEMES
    results = dict()
    for tr in tqdm(events.columns.get_level_values(level=peyes.constants.TRIAL_ID_STR).unique(), desc="Matching Events"):
        for gt_labeler in gt_labelers:
            trial_gt_events = events.xs((tr, gt_labeler), axis=1, level=[peyes.constants.TRIAL_ID_STR, peyes.constants.LABELER_STR])
            if trial_gt_events.size == 0:
                continue
            gt_min_iteration = np.nanmin(trial_gt_events.columns.get_level_values(peyes.constants.ITERATION_STR))
            gt_events = events[tr, gt_labeler, gt_min_iteration].dropna().values.flatten()
            if gt_events.size == 0:
                continue
            for pred_labeler in pred_labelers:
                pred_events_all_iters = events.xs(
                    (tr, pred_labeler), axis=1, level=[peyes.constants.TRIAL_ID_STR, peyes.constants.LABELER_STR]
                )
                for pred_it in pred_events_all_iters.columns.get_level_values(peyes.constants.ITERATION_STR).unique():
                    if (pred_labeler == gt_labeler) and (pred_it == gt_min_iteration):
                        continue
                    pred_events = pred_events_all_iters[pred_it].dropna().values.flatten()
                    if pred_events.size == 0:
                        continue
                    matches = dict()
                    for scheme_name, match_params in matching_schemes.items():
                        match_params[_MATCH_BY_STR] = match_params.get(_MATCH_BY_STR, scheme_name)
                        matches[scheme_name] = peyes.match(
                            gt_events, pred_events, allow_xmatch=allow_xmatch, **match_params
                        )
                    results[(tr, gt_labeler, pred_labeler, pred_it)] = matches
    results = pd.DataFrame.from_dict(results, orient="columns")
    results.index.names = [u.MATCHING_SCHEME_STR]
    results.columns.names = [peyes.constants.TRIAL_ID_STR, u.GT_STR, u.PRED_STR, peyes.constants.ITERATION_STR]
    return results



from peyes._DataModels.EventMatcher import OneToOneEventMatchesType as _OneToOneEventMatchesType

from peyes.match_metrics._get_features import get_features as features
from peyes.match_metrics._match_evaluation import (
    match_ratio, precision_recall_f1, false_alarm_rate, d_prime_and_criterion
)


def onset_difference(matches: _OneToOneEventMatchesType):
    return features(matches, 'onset', verbose=False)


def offset_difference(matches: _OneToOneEventMatchesType):
    return features(matches, 'offset', verbose=False)


def duration_difference(matches: _OneToOneEventMatchesType):
    return features(matches, 'duration', verbose=False)


def amplitude_difference(matches: _OneToOneEventMatchesType):
    return features(matches, 'amplitude', verbose=False)


def azimuth_difference(matches: _OneToOneEventMatchesType):
    return features(matches, 'azimuth', verbose=False)


def center_pixel_distance(matches: _OneToOneEventMatchesType):
    return features(matches, 'center_pixel_distance', verbose=False)


def time_overlap(matches: _OneToOneEventMatchesType):
    return features(matches, 'time_overlap', verbose=False)


def time_iou(matches: _OneToOneEventMatchesType):
    return features(matches, 'time_iou', verbose=False)


def time_l2(matches: _OneToOneEventMatchesType):
    return features(matches, 'time_l2', verbose=False)



import os
import time
from typing import List, Union, Dict, Optional

import numpy as np
import pandas as pd

import peyes
from peyes._DataModels.Detector import BaseDetector
from peyes._DataModels.UnparsedEventLabel import UnparsedEventLabelType, UnparsedEventLabelSequenceType

import analysis.utils as u
import analysis.process.preprocess as preprocess
import analysis.process.sample_metrics as sample_metrics
import analysis.process.temporal_alignment as temporal_alignment
import analysis.process.match_metrics as match_metrics


def full_pipeline(
        output_dir: str,
        dataset_name: str,
        detectors: List[BaseDetector],
        annotators: List[str] = None,
        num_iterations: int = 4,
        iterations_overwrite_label: Union[UnparsedEventLabelType, UnparsedEventLabelSequenceType] = 2,
        matching_schemes: Dict[str, Dict[str, Union[str, int, float]]] = preprocess.DEFAULT_MATCHING_SCHEMES,
        allow_xmatch: bool = False,
        pos_labels: Optional[Union[UnparsedEventLabelType, UnparsedEventLabelSequenceType]] = None,
        sample_sdt_average: str = "weighted",
        sample_dprime_correction: str = "loglinear",
        channel_max_difference: int = 250,
        channel_dprime_correction: str = "loglinear",
        verbose: bool = True
):
    start = time.time()
    ## Load dataset ##
    dataset = u.load_dataset(dataset_name, verbose=verbose)
    output_dir = os.path.join(output_dir, dataset_name)
    os.makedirs(output_dir, exist_ok=True)
    annotators = annotators or u.DATASET_ANNOTATORS[dataset_name]   # dataset's human annotators

    ## labels, metadata, events, matches ##
    try:
        labels = pd.read_pickle(os.path.join(output_dir, f"{peyes.constants.LABELS_STR}.pkl"))
        metadata = pd.read_pickle(os.path.join(output_dir, f"{peyes.constants.METADATA_STR}.pkl"))
        events = pd.read_pickle(os.path.join(output_dir, f"{peyes.constants.EVENTS_STR}.pkl"))
    except FileNotFoundError:
        labels, metadata, events = preprocess.detect_labels_and_events(
            dataset, detectors, annotators, num_iterations, iterations_overwrite_label, verbose
        )

        # replace non-fixation labels with 'undefined' label as HFC only includes fixation labels
        labels.to_pickle(os.path.join(output_dir, f"raw_{peyes.constants.LABELS_STR}.pkl"))
        labels = labels.replace({
            lbl: peyes.parse_label("undef", safe=True) for lbl in np.unique(labels.values)
            if lbl != peyes.parse_label("fixation", safe=False) or pd.isna(lbl)
        })

        labels.to_pickle(os.path.join(output_dir, f"{peyes.constants.LABELS_STR}.pkl"))
        metadata.to_pickle(os.path.join(output_dir, f"{peyes.constants.METADATA_STR}.pkl"))
        events.to_pickle(os.path.join(output_dir, f"{peyes.constants.EVENTS_STR}.pkl"))
    try:
        matches = pd.read_pickle(os.path.join(output_dir, f"{peyes.constants.MATCHES_STR}.pkl"))
    except FileNotFoundError:
        matches = preprocess.match_events(
            events, annotators, matching_schemes=matching_schemes, allow_xmatch=allow_xmatch
        )
        matches.to_pickle(os.path.join(output_dir, f"{peyes.constants.MATCHES_STR}.pkl"))

    ## Sample metrics ##
    sample_metrics_dir = os.path.join(output_dir, f"{peyes.constants.SAMPLE_STR}_{peyes.constants.METRICS_STR}")
    os.makedirs(sample_metrics_dir, exist_ok=True)

    # Sample SDT metrics
    sdt_sample_metrics_fullpath = os.path.join(
        sample_metrics_dir, u.get_filename_for_labels(
            pos_labels, suffix=f"{u.SDT_STR}_{peyes.constants.METRICS_STR}", extension="pkl"
        )
    )
    try:
        sdt_sample_metrics = pd.read_pickle(sdt_sample_metrics_fullpath)
    except FileNotFoundError:
        sdt_sample_metrics = sample_metrics.calculate_sdt_sample_metrics(
            labels, annotators, pos_labels=pos_labels, average=sample_sdt_average, correction=sample_dprime_correction
        )
        sdt_sample_metrics.to_pickle(sdt_sample_metrics_fullpath)

    # Sample Global metrics - only when pos_labels is None
    if pos_labels is None or (isinstance(pos_labels, list) and len(pos_labels) == 0):
        global_sample_metrics_fullpath = os.path.join(
            sample_metrics_dir, u.get_filename_for_labels(
                pos_labels, suffix=f"{u.GLOBAL_STR}_{peyes.constants.METRICS_STR}", extension="pkl"
            )
        )
        try:
            global_sample_metrics = pd.read_pickle(global_sample_metrics_fullpath)
        except FileNotFoundError:
            global_sample_metrics = sample_metrics.calculate_global_sample_metrics(labels, annotators)
            global_sample_metrics.to_pickle(global_sample_metrics_fullpath)
    else:
        global_sample_metrics = None

    ## Sample Channel metrics ##
    channel_metrics_dir = os.path.join(output_dir, peyes.constants.SAMPLES_CHANNEL_STR)
    os.makedirs(channel_metrics_dir, exist_ok=True)
    time_diffs_fullpath = os.path.join(
        channel_metrics_dir, u.get_filename_for_labels(pos_labels, suffix="timing_differences", extension="pkl")
    )
    try:
        time_diffs = pd.read_pickle(time_diffs_fullpath)
    except FileNotFoundError:
        time_diffs = temporal_alignment.timing_differences(
            labels, annotators, pos_labels=pos_labels, max_difference=channel_max_difference,
        )
        time_diffs.to_pickle(time_diffs_fullpath)
    channel_sdt_metrics_fullpath = os.path.join(
        channel_metrics_dir, u.get_filename_for_labels(pos_labels, suffix="sdt_metrics", extension="pkl")
    )
    try:
        channel_sdt_metrics = pd.read_pickle(channel_sdt_metrics_fullpath)
    except FileNotFoundError:
        channel_sdt_metrics = temporal_alignment.signal_detection_metrics(
            labels, np.arange(21), annotators, pos_labels=pos_labels, dprime_correction=channel_dprime_correction
        )
        channel_sdt_metrics.to_pickle(channel_sdt_metrics_fullpath)

    ## Match metrics ##
    match_metrics_dir = os.path.join(output_dir, f"{peyes.constants.MATCHES_STR}_{peyes.constants.METRICS_STR}")
    os.makedirs(match_metrics_dir, exist_ok=True)
    matched_features_fullpath = os.path.join(
        match_metrics_dir, u.get_filename_for_labels(labels=None, suffix="matched_features", extension="pkl")
    )
    try:
        matched_features = pd.read_pickle(matched_features_fullpath)
    except FileNotFoundError:
        matched_features = match_metrics.calculate_matched_features(matches)
        matched_features.to_pickle(matched_features_fullpath)
    matches_sdt_fullpath = os.path.join(
        match_metrics_dir, u.get_filename_for_labels(pos_labels, suffix="sdt_metrics", extension="pkl")
    )
    try:
        matches_sdt_metrics = pd.read_pickle(matches_sdt_fullpath)
    except FileNotFoundError:
        matches_sdt_metrics = match_metrics.calculate_event_sdt_measures(events, matches, pos_labels)
        matches_sdt_metrics.to_pickle(matches_sdt_fullpath)

    elapsed = time.time() - start
    if verbose:
        print(f"Finished in {elapsed:.2f}s")
    return (
        dataset,
        labels,
        metadata,
        events,
        matches,
        sdt_sample_metrics,
        global_sample_metrics,
        time_diffs,
        channel_sdt_metrics,
        matched_features,
        matches_sdt_metrics,
    )



import os

import pandas as pd
import plotly.io as pio

import peyes
import analysis.utils as u
from analysis._default_values._helpers import DATASET_NAME, STIMULUS_TYPE, PROCESSED_DATA_DIR, FIGURES_DIR

pio.renderers.default = "browser"

######################
# Load Events Data

stim_trial_ids = u.get_trials_for_stimulus_type(DATASET_NAME, STIMULUS_TYPE)

all_events = pd.read_pickle(os.path.join(PROCESSED_DATA_DIR, DATASET_NAME, "events.pkl"))
all_events = all_events.xs(1, level=peyes.constants.ITERATION_STR, axis=1)
all_events = all_events.loc[:, all_events.columns.get_level_values(peyes.constants.TRIAL_ID_STR).isin(stim_trial_ids)]
all_events = all_events.dropna(axis=0, how="all")

all_labelers = all_events.columns.get_level_values(peyes.constants.LABELER_STR).unique()
events_by_labelers = {
    lblr: all_events.xs(lblr, level=peyes.constants.LABELER_STR, axis=1).stack().dropna() for lblr in all_labelers
}

######################
# Fixation Features Comparison

fixations_comparison_figure = peyes.visualize.feature_comparison(
    [
        peyes.constants.DURATION_STR,
        peyes.constants.AMPLITUDE_STR,
        peyes.constants.PEAK_VELOCITY_STR,
        peyes.constants.MEDIAN_VELOCITY_STR,
        peyes.constants.COUNT_STR,
     ],
    *[vals[vals.apply(lambda e: e.label == 1)] for vals in events_by_labelers.values()],
    labels=events_by_labelers.keys(),
    title="Fixation Features Comparison",
)
fixations_comparison_figure.update_layout(
    width=1440, height=900
)
fixations_comparison_figure.show()

######################
# Saccade Features Comparison

saccades_comparison_figure = peyes.visualize.feature_comparison(
    [
        peyes.constants.DURATION_STR,
        peyes.constants.AMPLITUDE_STR,
        peyes.constants.PEAK_VELOCITY_STR,
        peyes.constants.MEDIAN_VELOCITY_STR,
        peyes.constants.COUNT_STR,
     ],
    *[vals[vals.apply(lambda e: e.label == 2)] for vals in events_by_labelers.values()],
    labels=events_by_labelers.keys(),
    title="Saccade Features Comparison",
)
saccades_comparison_figure.update_layout(
    width=1440, height=900
)
saccades_comparison_figure.show()

######################
# Save Figures

peyes.visualize.save_figure(
    fixations_comparison_figure, "features_comparison-fixation", FIGURES_DIR, as_png=True, as_html=False, as_json=False
)

peyes.visualize.save_figure(
    saccades_comparison_figure, "features_comparison-saccade", FIGURES_DIR, as_png=True, as_html=False, as_json=False
)



#
# The following scripts are meant to compare velocity thresholds of the I-DT algorithm:
#   - Andersson et al. (2017): 2.7 deg + 55 ms
#   - Salvucci & Goldberg (2000): 1 deg + 100 ms
#   - Birawo & Kasprowski (2022): 0.1 deg + 10 ms - not analyzed
#
# We evaluate detection performance compared to human coders on the lund2013+image dataset (only image stimuli).
# The following metrics are used:
#   1) overall sample-level agreement - balanced accuracy, Cohen's Kappa, MCC, 1-NLD
#   2) fixation onset/offset discriminability (d')
#   3) saccade onset/offset discriminability (d')
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ###

import pandas as pd
import plotly.io as pio

import peyes

import analysis._parameter_exploration._helpers as h

pio.renderers.default = "browser"

COMPARED_DETECTORS = [det for det in h.DETECTORS.keys() if "idt" in det.lower()]

# %%
# SAMPLE-LEVEL AGREEMENT (GLOBAL MEASURES)

import analysis.statistics.sample_metrics as sm

# Load Data
global_metrics = sm.load_global_metrics(
    h.DATASET_NAME, h.PROCESSED_DATA_DIR, stimulus_type=h.STIMULUS_TYPE, metric=None, iteration=1,
)
global_metrics = pd.concat(
    [global_metrics.xs(det, axis=1, level='pred', drop_level=False) for det in COMPARED_DETECTORS], axis=1
).drop(index=peyes.constants.ACCURACY_STR, inplace=False)    # drop Accuracy metric

global_statistics, global_pvalues, global_Ns = sm.wilcoxon(
    global_metrics, [h.GT1, h.GT2], method='exact'
)

global_distribution_figure = sm.global_metrics_distributions_figure(global_metrics, h.GT1, h.GT2,)
global_distribution_figure.show()

print("Salvucci *significantly* outperforms Andersson! (except for 1-NLD)")

# %%
# ONSET & OFFSET DETECTION (FIXATIONS)

import analysis.statistics.channel_sdt as ch_sdt

LABEL = 1       # EventLabelEnum.FIXATION.value
THRESHOLD = 5   # samples

# Load Data
fixation_csdt_metrics = ch_sdt.load(
    dataset_name=h.DATASET_NAME,
    output_dir=h.PROCESSED_DATA_DIR,
    label=LABEL,
    stimulus_type=h.STIMULUS_TYPE,
    channel_type=None
)
fixation_csdt_metrics = pd.concat(
    [fixation_csdt_metrics.xs(det, axis=1, level='pred', drop_level=False) for det in COMPARED_DETECTORS], axis=1
)
fixation_csdt_metrics.drop(index=['P', 'PP', 'TP', 'N'], level=peyes.constants.METRIC_STR, inplace=True)    # drop irrelevant metrics

# calc stats
fix_onset_w_stat, fix_onset_pvalues, fix_onset_Ns = ch_sdt.wilcoxon(
    fixation_csdt_metrics, "onset", THRESHOLD, [h.GT1, h.GT2], method='exact',
)

fix_offset_w_stat, fix_offset_pvalues, fix_offset_Ns = ch_sdt.wilcoxon(
    fixation_csdt_metrics, "offset", THRESHOLD, [h.GT1, h.GT2], method='exact',
)

# show figures
fixation_dprime_figure = ch_sdt.multi_channel_figure(
    fixation_csdt_metrics,
    metric=peyes.constants.D_PRIME_STR,
    yaxis_title=r"$d'$", show_other_gt=True, show_err_bands=True
)
fixation_dprime_figure.update_layout(width=1400, height=500,)
fixation_dprime_figure.show()

print("Salvucci *significantly* outperforms Andersson on fixations onsets, but not fixation offsets.")

# %%
# ONSET & OFFSET DETECTION (SACCADES)

import analysis.statistics.channel_sdt as ch_sdt

LABEL = 2       # EventLabelEnum.SACCADE.value
THRESHOLD = 5   # samples

# Load Data
saccade_csdt_metrics = ch_sdt.load(
    dataset_name=h.DATASET_NAME,
    output_dir=h.PROCESSED_DATA_DIR,
    label=LABEL,
    stimulus_type=h.STIMULUS_TYPE,
    channel_type=None
)
saccade_csdt_metrics = pd.concat(
    [saccade_csdt_metrics.xs(det, axis=1, level='pred', drop_level=False) for det in COMPARED_DETECTORS], axis=1
)
saccade_csdt_metrics.drop(index=['P', 'PP', 'TP', 'N'], level=peyes.constants.METRIC_STR, inplace=True)    # drop irrelevant metrics

# calc stats
sac_onset_w_stat, sac_onset_pvalues, sac_onset_Ns = ch_sdt.wilcoxon(
    saccade_csdt_metrics, "onset", THRESHOLD, [h.GT1, h.GT2], method='exact',
)

sac_offset_w_stat, sac_offset_pvalues, sac_offset_Ns = ch_sdt.wilcoxon(
    saccade_csdt_metrics, "offset", THRESHOLD, [h.GT1, h.GT2], method='exact',
)

# show figures
saccade_dprime_figure = ch_sdt.multi_channel_figure(
    saccade_csdt_metrics,
    metric=peyes.constants.D_PRIME_STR,
    yaxis_title=r"$d'$", show_other_gt=True, show_err_bands=True
)
saccade_dprime_figure.update_layout(width=1400, height=500,)
saccade_dprime_figure.show()

print("Salvucci *significantly* outperforms Andersson on saccades, but only for offsets")



from typing import Union

import numpy as np

from peyes._DataModels.Event import EventSequenceType
from peyes._DataModels.UnparsedEventLabel import UnparsedEventLabelSequenceType

from peyes._base.create import create_boolean_channel
from peyes._utils.vector_utils import pair_boolean_arrays


def onset_differences(
        ground_truth: Union[UnparsedEventLabelSequenceType, EventSequenceType],
        prediction: Union[UnparsedEventLabelSequenceType, EventSequenceType],
        max_diff: int = 25,
        sampling_rate: float = None,
        min_num_samples=None,
) -> np.ndarray:
    """
    1. Converts the label/event sequences to a boolean array indicating event onsets (MNE-style event channel).
    2. Matches between `True` indices (event onsets) of the two channels such that the difference between their indices
        is minimal.
    3. Calculates the difference between the matched indices, in samples.

    :param ground_truth: array-like of ground-truth labels or Event objects
    :param prediction: array-like of predicted labels or Event objects
    :param max_diff: maximal difference between matched onsets/offsets (in samples); default is 25 samples
    :param sampling_rate: sampling rate of the channels; only used if `ground_truth` or `prediction` are Event objects
    :param min_num_samples: if not None, marks the minimal the number of samples in the channels; only used if
        `ground_truth` or `prediction` are Event objects
    :return: array of timing differences between matched onsets (differences are in sample units)
    """
    return timing_differences(ground_truth, prediction, "onset", max_diff, sampling_rate, min_num_samples)


def offset_differences(
        ground_truth: Union[UnparsedEventLabelSequenceType, EventSequenceType],
        prediction: Union[UnparsedEventLabelSequenceType, EventSequenceType],
        max_diff: int = 25,
        sampling_rate: float = None,
        min_num_samples=None,
) -> np.ndarray:
    """
    1. Converts the label/event sequences to a boolean array indicating event offsets (MNE-style event channel).
    2. Matches between `True` indices (event offsets) of the two channels such that the difference between their indices
        is minimal.
    3. Calculates the difference between the matched indices, in samples.

    :param ground_truth: array-like of ground-truth labels or Event objects
    :param prediction: array-like of predicted labels or Event objects
    :param max_diff: maximal difference between matched onsets/offsets (in samples); default is 25 samples
    :param sampling_rate: sampling rate of the channels; only used if `ground_truth` or `prediction` are Event objects
    :param min_num_samples: if not None, marks the minimal the number of samples in the channels; only used if
        `ground_truth` or `prediction` are Event objects
    :return: array of timing differences between matched offsets (differences are in sample units)
    """
    return timing_differences(ground_truth, prediction, "offset", max_diff, sampling_rate, min_num_samples)


def timing_differences(
        ground_truth: Union[UnparsedEventLabelSequenceType, EventSequenceType],
        prediction: Union[UnparsedEventLabelSequenceType, EventSequenceType],
        channel_type: str,
        max_diff: int = 25,
        sampling_rate: float = None,
        min_num_samples=None,
) -> np.ndarray:
    """
    Converts each of the label/event sequences to a boolean channel, matches between `True` indices of the two channels such
    that the difference between their indices is minimal, and calculates the difference between the matched indices, in
    samples. Returns an array of timing differences between matched events.

    :param ground_truth: array-like of ground-truth labels or Event objects
    :param prediction: array-like of predicted labels or Event objects
    :param channel_type: either 'onset' or 'offset'
    :param max_diff: maximal difference between matched onsets/offsets (in samples); default is 25 samples
    :param sampling_rate: sampling rate of the channels; only used if `ground_truth` or `prediction` are Event objects
    :param min_num_samples: if not None, marks the minimal the number of samples in the channels; only used if
        `ground_truth` or `prediction` are Event objects

    :return: array of timing differences between matched onsets/offsets (differences are in samples)
    """
    gt_channel = create_boolean_channel(
        data=ground_truth, channel_type=channel_type, sampling_rate=sampling_rate, min_num_samples=min_num_samples
    )
    pred_channel = create_boolean_channel(
        data=prediction, channel_type=channel_type, sampling_rate=sampling_rate, min_num_samples=min_num_samples
    )
    matched_idxs = pair_boolean_arrays(gt_channel, pred_channel)
    diffs = np.diff(matched_idxs, axis=1).flatten()
    diffs = diffs[np.abs(diffs) <= max_diff]
    return diffs



from typing import Dict

import numpy as np
import pandas as pd
import plotly.graph_objects as go
from plotly.subplots import make_subplots

import peyes._utils.constants as cnst
import peyes._utils.visualization_utils as vis_utils
from peyes._utils.event_utils import parse_label
from peyes.event_metrics import features_by_labels
from peyes import summarize_events

from peyes._DataModels.Event import EventSequenceType
from peyes._DataModels.EventLabelEnum import EventLabelEnum

__INLIERS_STR, __OUTLIERS_STR = "inliers", "outliers"


def event_summary(events: EventSequenceType, **kwargs) -> go.Figure:
    """
    Creates a summary figure of the provided events. The figure contains four panels:
        1. A bar plot of the count of each event label.
        2. A violin plot showing the distribution of durations, per event label.
        3. A violin plot showing the distribution of amplitudes, per event label.
        4. A violin plot showing the distribution of peak velocities, per event label.
        5. A violin plot showing the distribution of areas of the ellipse bound each event, per event label.

    :param events: list (or other sequence) of events.
    :keyword label_colors: A dictionary mapping event labels to their respective colors. If a label is missing, the
        default color is used.
    :keyword show_outliers: whether to show outliers in the violin plots; default is False.
    :keyword outlier_opacity: if `show_outliers` is true, this controls the opacity of the outliers; default is 0.5.
    :keyword title: the title of the figure; default is "Event Summary".

    :return: the figure.
    """
    inlier_features = features_by_labels(list(filter(lambda e: not e.is_outlier, events)))
    outlier_features = features_by_labels(list(filter(lambda e: e.is_outlier, events)))
    assert inlier_features.index.equals(outlier_features.index)
    event_labels = [parse_label(l) for l in inlier_features.index]
    label_colors = vis_utils.get_label_colormap(kwargs.get("label_colors", None))
    show_outliers = kwargs.get("show_outliers", False)
    subplots = {
        cnst.COUNT_STR: '# instances',
        cnst.DURATION_STR: 'ms',
        cnst.AMPLITUDE_STR: 'deg',
        cnst.PEAK_VELOCITY_STR: 'deg/s',
        cnst.ELLIPSE_AREA_STR: 'deg^2',
    }
    fig = make_subplots(
        cols=1, rows=len(subplots), shared_xaxes=True,
        subplot_titles=list(map(lambda met: met.replace('_', ' ').title(), subplots.keys()))
    )
    for r, (metric, measure) in enumerate(subplots.items()):
        fig.update_yaxes(col=1, row=r + 1, title_text=measure)
        for evnt in event_labels:
            outlier_trace = None
            color = f"rgb{label_colors[evnt]}"
            if metric == cnst.COUNT_STR:
                inlier_trace = go.Bar(
                    x0=evnt.name,
                    y=[inlier_features.loc[evnt, metric]],
                    marker_color=color,
                    name=evnt.name, legendgroup=evnt.name, showlegend=True,
                )
                if show_outliers:
                    outlier_trace = go.Bar(
                        x0=evnt.name,
                        y=[outlier_features.loc[evnt, metric]],
                        marker_color=color, opacity=kwargs.get("outlier_opacity", 0.5),
                        name=f"{evnt.name} ({__OUTLIERS_STR})", legendgroup=evnt.name, showlegend=False,
                    )
            else:
                inlier_trace = go.Violin(
                    x0=evnt.name,
                    y=inlier_features.loc[evnt, metric],
                    fillcolor=color, spanmode='hard', showlegend=False,
                    side='positive' if show_outliers else None,
                    box_visible=True, meanline_visible=True, line_color='black',
                    name=evnt.name, legendgroup=evnt.name,
                )
                if show_outliers:
                    outlier_trace = go.Violin(
                        x0=evnt.name,
                        y=outlier_features.loc[evnt, metric],
                        fillcolor=color, spanmode='hard', showlegend=False, side='negative',
                        opacity=kwargs.get("outlier_opacity", 0.5),
                        box_visible=True, meanline_visible=True, line_color='black',
                        name=f"{evnt.name} ({__OUTLIERS_STR})", legendgroup=evnt.name,
                    )
            fig.add_trace(inlier_trace, col=1, row=r+1)
            if show_outliers and outlier_trace is not None:
                fig.add_trace(outlier_trace, col=1, row=r+1)
    fig.update_layout(
        title=kwargs.get("title", f"{cnst.EVENT_STR.title()} Summary"),
        barmode='stack',
        violinmode='overlay',
        violingap=0,
        violingroupgap=0.1,
    )
    return fig


def fixation_summary(fixations: EventSequenceType, **kwargs) -> go.Figure:
    """
    Creates a 2×2 figure with the following subplots:
        1. A violin plot showing the distribution of fixation durations (ms)
        2. A violin plot showing the distribution of fixation areas (deg^2)
        3. A violin plot showing the distribution of peak velocities (deg/s)
        4. A scatter plot showing the distribution of fixation centers (x-y coordinates)

    :param fixations: array-like of fixation events.

    :keyword show_outliers: whether to show outliers in the violin plots; default is False.
    :keyword inlier_color: the color of the inliers.
    :keyword outlier_color: the color of the outliers.
    :keyword outlier_opacity: the opacity of the outliers; default is 0.5.
    :keyword marker_size: the size of the markers in the scatter plot; default is 5.
    :keyword title: the title of the figure; default is "Fixation Summary".

    :return: the figure.
    """

    fixations = list(filter(lambda e: e.label == EventLabelEnum.FIXATION, fixations))   # filter out non-fixation events
    subplots = {
        cnst.DURATION_STR: 'ms', cnst.ELLIPSE_AREA_STR: 'deg^2',
        cnst.PEAK_VELOCITY_STR: 'deg/s', cnst.CENTER_PIXEL_STR: 'x coordinate',
    }
    return _create_single_event_figure(fixations, subplots, **kwargs)


def saccade_summary(saccades: EventSequenceType, **kwargs) -> go.Figure:
    """
    Creates a 2×2 figure with the following subplots:
        1. A violin plot showing the distribution of saccade durations (ms)
        2. A violin plot showing the distribution of saccade amplitudes (deg)
        3. A violin plot showing the distribution of peak velocities (deg/s)
        4. A polar histogram showing the distribution of saccade azimuths (deg)

    :param saccades: array-like of saccade events.

    :keyword show_outliers: whether to show outliers in the violin plots; default is False.
    :keyword inlier_color: the color of the inliers.
    :keyword outlier_color: the color of the outliers.
    :keyword outlier_opacity: the opacity of the outliers; default is 0.5.
    :keyword azimuth_nbins: the number of bins in the azimuth histogram; default is 16.
    :keyword title: the title of the figure; default is "Saccade Summary".

    :return: the figure.
    """
    saccades = list(filter(lambda e: e.label == EventLabelEnum.SACCADE, saccades))   # filter out non-saccade events
    subplots = {
        cnst.DURATION_STR: 'ms', cnst.AMPLITUDE_STR: 'deg',
        cnst.PEAK_VELOCITY_STR: 'deg/s', cnst.AZIMUTH_STR: 'deg',
    }
    return _create_single_event_figure(saccades, subplots, **kwargs)


def _create_single_event_figure(
        events: EventSequenceType,
        subplots: Dict[str, str],
        **kwargs,
) -> go.Figure:
    """
    Creates a figure with multiple subplots, each showing a different metric of the provided events.

    :param events: array-like of events. All events should be of the same type
    :param subplots: a dictionary mapping metric names to their respective measures (e.g. `peak velocity` -> `deg/s`)

    :keyword num_cols: the number of columns in the figure; default is 2.
        Number of rows is calculated as np.ceil(len(subplots) / num_cols).
    :keyword inlier_color: the color of the inliers; default is the color of the event label.
    :keyword show_outliers: whether to show outliers in the violin plots; default is False.
    :keyword outlier_color: the color of the outliers; default is the color of the event label.
    :keyword outlier_opacity: the opacity of the outliers; default is 0.5.
    :keyword marker_size: the size of the markers in scatter plots; default is 5.
    :keyword azimuth_nbins: the number of bins in the azimuth histogram; default is 16.
    :keyword title: the title of the figure; default is "{Event-Label} Summary" (e.g. "Saccade Summary").

    :return: the figure.
    """
    event_label = events[0].label
    assert all(e.label == event_label for e in events), "All events must have the same label."
    summary_df = summarize_events(events)
    inlier_df, outlier_df = summary_df[~summary_df[cnst.IS_OUTLIER_STR]], summary_df[summary_df[cnst.IS_OUTLIER_STR]]

    # create figure:
    num_cols = kwargs.get("num_cols", 2)
    num_rows = np.ceil(len(subplots) / num_cols).astype(int).item()
    num_subfigs = num_rows * num_cols
    subplot_titles = list(map(lambda met: met.replace('_', ' ').title(), subplots.keys())) + [None] * (num_subfigs - len(subplots))
    specs = [{}] * len(subplots) + [None] * (num_subfigs - len(subplots))
    if cnst.AZIMUTH_STR in subplots.keys():
        specs[list(subplots.keys()).index(cnst.AZIMUTH_STR)] = {'type': 'polar'}
    fig = make_subplots(
        rows=num_rows, cols=num_cols,
        subplot_titles=subplot_titles,
        specs=np.array(specs).reshape(num_rows, num_cols).tolist(),
    )

    # populate subplots with traces:
    for i, (metric, measure) in enumerate(subplots.items()):
        r, c = divmod(i, num_cols)
        fig.add_trace(row=r + 1, col=c + 1, trace=__create_single_event_trace(
            data=inlier_df[metric],
            metric=metric,
            is_inlier=True,
            color=kwargs.get(
                "inlier_color", f"rgb{vis_utils.get_label_colormap(None)[event_label]}"
            ),
            marker_size=kwargs.get("marker_size", 5),
            nbins_polar=kwargs.get(f"{cnst.AZIMUTH_STR}_nbins", 16),
            show_legend=i == 1,  # show legend on top-right figure only
        ))
        if kwargs.get("show_outliers", False):
            fig.add_trace(row=r + 1, col=c + 1, trace=__create_single_event_trace(
                data=outlier_df[metric],
                metric=metric,
                is_inlier=False,
                color=kwargs.get(
                    "outlier_color", f"rgb{vis_utils.get_label_colormap(None)[event_label]}"
                ),
                outlier_opacity=kwargs.get("outlier_opacity", 0.5),
                marker_size=kwargs.get("marker_size", 5),
                nbins_polar=kwargs.get(f"{cnst.AZIMUTH_STR}_nbins", 16),
                show_legend=i == 1,  # show legend on top-right figure only
            ))
        fig.update_xaxes(title_text=measure, row=r + 1, col=c + 1)
        if metric == cnst.CENTER_PIXEL_STR:
            fig.update_yaxes(title_text="y coordinate", autorange='reversed', row=r + 1, col=c + 1)

    # adjust layout:
    fig.update_layout(
        title=kwargs.get("title", f"{event_label.name.title()} Summary"),
        violinmode='overlay',
        violingap=0,
        violingroupgap=0,
    )
    return fig


def __create_single_event_trace(
        data: pd.Series,
        metric: str,
        is_inlier: bool,
        color: str,
        marker_size: int,
        nbins_polar: int,
        outlier_opacity: float = 1,
        show_legend: bool = False,
):
    name = __INLIERS_STR if is_inlier else __OUTLIERS_STR
    opacity = 1 if is_inlier else outlier_opacity
    if metric == cnst.CENTER_PIXEL_STR:
        px = pd.DataFrame(data.to_list(), columns=[cnst.X, cnst.Y])  # data is a 2-tuple of x-y coordinates
        return go.Scatter(
            x=px[cnst.X], y=px[cnst.Y], mode="markers", name=name, legendgroup=name,
            marker=dict(size=marker_size, color=color, opacity=opacity), showlegend=show_legend,
        )
    if metric == cnst.AZIMUTH_STR:
        half_bin = 360 / nbins_polar / 2
        corrected_azimuths = (data.values + half_bin) % 360
        edges = np.linspace(0, 360, nbins_polar + 1, endpoint=True)
        counts, _ = np.histogram(corrected_azimuths, bins=edges)
        return go.Barpolar(
            r=counts, name=name, legendgroup=name, showlegend=show_legend,
            opacity=opacity, marker=dict(color=color),
        )
    if metric in {
        cnst.DURATION_STR, cnst.AMPLITUDE_STR, cnst.PEAK_VELOCITY_STR, cnst.ELLIPSE_AREA_STR
    }:
        violin_side = 'positive' if is_inlier else 'negative'
        return go.Violin(
            x=data, side=violin_side, showlegend=show_legend,
            name=name, legendgroup=name, scalegroup=metric, fillcolor=color, opacity=opacity,
            box_visible=True, meanline_visible=True, line_color='black',
            spanmode='hard', orientation='h', points=False,
        )
    raise ValueError(f"Invalid metric: {metric}")



from typing import Union, Optional

from peyes._utils import constants as _cnst
from peyes._DataModels.EventLabelEnum import EventLabelSequenceType as _EventLabelSequenceType
from peyes._DataModels.UnparsedEventLabel import UnparsedEventLabelType as _UnparsedEventLabelType
from peyes._DataModels.UnparsedEventLabel import UnparsedEventLabelSequenceType as _UnparsedEventLabelSequenceType

from peyes.sample_metrics._counts_and_matrices import label_counts, transition_matrix, confusion_matrix
from peyes.sample_metrics._calculate_metrics import calculate


def accuracy(ground_truth: _EventLabelSequenceType, prediction: _EventLabelSequenceType,) -> float:
    return calculate(ground_truth, prediction, _cnst.ACCURACY_STR)


def balanced_accuracy(ground_truth: _EventLabelSequenceType, prediction: _EventLabelSequenceType,) -> float:
    return calculate(ground_truth, prediction, _cnst.BALANCED_ACCURACY_STR)


def cohen_kappa(ground_truth: _EventLabelSequenceType, prediction: _EventLabelSequenceType,) -> float:
    return calculate(ground_truth, prediction, _cnst.COHENS_KAPPA_STR)


def mcc(ground_truth: _EventLabelSequenceType, prediction: _EventLabelSequenceType,) -> float:
    return calculate(ground_truth, prediction, _cnst.MCC_STR)


def complement_nld(ground_truth: _EventLabelSequenceType, prediction: _EventLabelSequenceType) -> float:
    """
    Calculates the complement of normalized Levenshtein distance (1-NLD) between two sequences.
    Levenshtein distance is the number of edits required to transform one sequence into another, and its normalization
    is with respect to the size of the GT (equivalent to the WER measure used in speech recognition). This normalization
    may result in a value exceeding 1 if the predicted sequence is longer than the GT sequence, meaning the complement
    (1-NLD) may be negative.
    For more information on using 1-NLD in eye-tracking, see https://doi.org/10.3758/s13428-021-01763-7
    """
    return calculate(ground_truth, prediction, _cnst.COMPLEMENT_NLD_STR)


def precision(
        ground_truth: _EventLabelSequenceType,
        prediction: _EventLabelSequenceType,
        pos_labels: Optional[Union[_UnparsedEventLabelType, _UnparsedEventLabelSequenceType]],
        average: str = "weighted",
) -> float:
    """
    Calculate the precision of the prediction sequence compared to the ground truth.
    :param ground_truth: the ground truth sequence of labels
    :param prediction: the predicted sequence of labels
    :param pos_labels: the positive label(s) to consider. Other labels (if exist) are considered negative.
    :param average: the averaging strategy for precision. default is "weighted"
    """
    return calculate(ground_truth, prediction, _cnst.PRECISION_STR, pos_labels=pos_labels, average=average)


def recall(
        ground_truth: _EventLabelSequenceType,
        prediction: _EventLabelSequenceType,
        pos_labels: Optional[Union[_UnparsedEventLabelType, _UnparsedEventLabelSequenceType]],
        average: str = "weighted",
) -> float:
    """
    Calculate the recall of the prediction sequence compared to the ground truth.
    :param ground_truth: the ground truth sequence of labels
    :param prediction: the predicted sequence of labels
    :param pos_labels: the positive label(s) to consider. Other labels (if exist) are considered negative.
    :param average: the averaging strategy for recall. default is "weighted"
    """
    return calculate(ground_truth, prediction, _cnst.RECALL_STR, pos_labels=pos_labels, average=average)


def f1_score(
        ground_truth: _EventLabelSequenceType,
        prediction: _EventLabelSequenceType,
        pos_labels: Optional[Union[_UnparsedEventLabelType, _UnparsedEventLabelSequenceType]],
        average: str = "weighted",
) -> float:
    """
    Calculate the F1-score of the prediction sequence compared to the ground truth.
    :param ground_truth: the ground truth sequence of labels
    :param prediction: the predicted sequence of labels
    :param pos_labels: the positive label(s) to consider. Other labels (if exist) are considered negative.
    :param average: the averaging strategy for F1-score. default is "weighted"
    """
    return calculate(ground_truth, prediction, _cnst.F1_STR, pos_labels=pos_labels, average=average)


def d_prime(
        ground_truth: _EventLabelSequenceType,
        prediction: _EventLabelSequenceType,
        pos_labels: Union[_UnparsedEventLabelType, _UnparsedEventLabelSequenceType],
        correction: str = "loglinear",
) -> float:
    """
    Calculates the discriminability (d-prime) of the prediction sequence compared to the ground truth.
    :param ground_truth: the ground truth sequence of labels
    :param prediction: the predicted sequence of labels
    :param pos_labels: the positive label(s) to consider. Other labels (if exist) are considered negative.
    :param correction: the correction method for floor/ceiling effects on the hit-rate and/or false-alarm rate.
        default is "loglinear".
        See information on correction methods at https://stats.stackexchange.com/a/134802/288290.
        See implementation details at https://lindeloev.net/calculating-d-in-python-and-php/.
    """
    return calculate(ground_truth, prediction, _cnst.D_PRIME_STR, pos_labels=pos_labels, correction=correction)


def criterion(
        ground_truth: _EventLabelSequenceType,
        prediction: _EventLabelSequenceType,
        pos_labels: Union[_UnparsedEventLabelType, _UnparsedEventLabelSequenceType],
        correction: str = "loglinear",
) -> float:
    """
    Calculates the criterion of the prediction sequence compared to the ground truth.
    :param ground_truth: the ground truth sequence of labels
    :param prediction: the predicted sequence of labels
    :param pos_labels: the positive label(s) to consider. Other labels (if exist) are considered negative.
    :param correction: the correction method for floor/ceiling effects on the hit-rate and/or false-alarm rate.
        default is "loglinear".
        See information on correction methods at https://stats.stackexchange.com/a/134802/288290.
        See implementation details at https://lindeloev.net/calculating-d-in-python-and-php/.
    """
    return calculate(
        ground_truth, prediction, _cnst.CRITERION_STR, pos_labels=pos_labels, correction=correction
    )



import os
from typing import Union, Tuple, Dict, Sequence

import cv2
import numpy as np
import plotly.graph_objects as go
import plotly.express as px
from plotly.subplots import make_subplots

from peyes._DataModels.EventLabelEnum import EventLabelEnum as _EventLabelEnum
from peyes._DataModels.UnparsedEventLabel import UnparsedEventLabelType as _UnparsedEventLabelType

ColorType = Union[str, Tuple[int, int, int]]
LabelColormapType = Dict[_UnparsedEventLabelType, ColorType]

_DISCRETE_COLORMAP = px.colors.qualitative.Dark24
_DEFAULT_COLORMAP = {
    _EventLabelEnum.UNDEFINED: "#dddddd",
    _EventLabelEnum.FIXATION: "#1f78b4",
    _EventLabelEnum.SACCADE: "#33a02c",
    _EventLabelEnum.PSO: "#b2df8a",
    _EventLabelEnum.SMOOTH_PURSUIT: "#fb9a99",
    _EventLabelEnum.BLINK: "#222222",
    "ra": _DISCRETE_COLORMAP[0],
    "rz": _DISCRETE_COLORMAP[0],
    'ih': _DISCRETE_COLORMAP[0],
    "mn": _DISCRETE_COLORMAP[1],
    'dn': _DISCRETE_COLORMAP[1],
    "ivt": _DISCRETE_COLORMAP[2],
    "ivtdetector": _DISCRETE_COLORMAP[2],
    "ivvt": _DISCRETE_COLORMAP[3],
    "ivvtdetector": _DISCRETE_COLORMAP[3],
    "idt": _DISCRETE_COLORMAP[4],
    "idtdetector": _DISCRETE_COLORMAP[4],
    "idvt": _DISCRETE_COLORMAP[5],
    "idvtdetector": _DISCRETE_COLORMAP[5],
    "engbert": _DISCRETE_COLORMAP[6],
    "engbertdetector": _DISCRETE_COLORMAP[6],
    "nh": _DISCRETE_COLORMAP[7],
    "nhdetector": _DISCRETE_COLORMAP[7],
    "remodnav": _DISCRETE_COLORMAP[8],
    "remodnavdetector": _DISCRETE_COLORMAP[8],
}


def save_figure(
        fig: go.Figure, fig_name: str, output_dir: str,
        width: int = None, height: int = None, scale: float = 1,
        as_json: bool = False, as_html: bool = False, as_png: bool = False
):
    os.makedirs(output_dir, exist_ok=True)
    if as_json:
        fig.write_json(os.path.join(output_dir, f"{fig_name}.json"))
    if as_html:
        fig.write_html(os.path.join(output_dir, f"{fig_name}.html"))
    if as_png:
        fig.write_image(os.path.join(output_dir, f"{fig_name}.png"), scale=scale, width=width, height=height)


def create_image(
        resolution: Tuple[int, int],
        image: np.ndarray = None,
        alpha: float = 1,
        color_format: str = "BGR",
        default_color: ColorType = (0, 0, 0),
):
    """
    Creates an image in BGR format with the specified resolution. If a background image is not provided, an image with
    the specified background color will be created.

    :param resolution: tuple of (width, height) in pixels
    :param image: background image (numpy array)
    :param alpha: alpha (opacity) value of the background image, range [0, 1]. Only used if input image doesn't have
    an alpha channel (i.e. `color_format` is not RGBA or BGRA). Default is 1 (opaque).
    :param color_format: color format of the background image (RGB/GRAY/BGR). Default is BGR.
    :param default_color: background color (RGB tuple or hex string). Default is black.

    :return: numpy array of the image
    """
    if not resolution or len(resolution) != 2 or resolution[0] <= 0 or resolution[1] <= 0:
        raise ValueError("resolution must be a tuple of two positive integers")
    if image.ndim != 4 and (alpha < 0 or alpha > 1):
        raise ValueError("bg_alpha must be in the range [0, 1]")

    # If bg_image is not provided or invalid, create an image with the specified RGB background color
    if image is None or image.size == 0 or image.ndim not in (2, 3, 4):
        default_color = to_rgb(default_color)
        image = np.full((resolution[1], resolution[0], 3), default_color, dtype=np.uint8)
        color_format = "RGB"

    # Convert the background image to BGRA format
    if color_format.upper() == "BGRA":
        bg = image
    elif color_format.upper() == "BGR":
        bg = cv2.cvtColor(image, cv2.COLOR_BGR2BGRA)
        bg[:, :, 3] = round(alpha * 255)
    elif color_format.upper() == "RGBA":
        bg = cv2.cvtColor(image, cv2.COLOR_RGBA2BGRA)
    elif color_format.upper() == "RGB":
        bg = cv2.cvtColor(image, cv2.COLOR_RGB2BGRA)
        bg[:, :, 3] = round(alpha * 255)
    elif color_format.upper() == "GRAY" or color_format.upper() == "GREY":
        bg = cv2.cvtColor(image, cv2.COLOR_GRAY2BGRA)
        bg[:, :, 3] = round(alpha * 255)
    else:
        raise ValueError(f"Invalid color format: {color_format}")
    return cv2.resize(bg, resolution)


def get_label_colormap(
        label_colors: LabelColormapType = None
):
    """
    Returns a dictionary mapping event labels to RGB colors.
    If a custom mapping is provided, it will override the default colors for the labels that are present, and use the
    default colors for the rest.
    """
    default_colors_rgb = {k: to_rgb(v) for k, v in _DEFAULT_COLORMAP.items()}
    if label_colors is None:
        return default_colors_rgb
    event_colors_rgb = {k: to_rgb(v) if isinstance(v, str) else v for k, v in label_colors.items()}
    return {**default_colors_rgb, **event_colors_rgb}


def to_rgb(color: ColorType) -> Tuple[int, int, int]:
    """
    Converts a hex color code to an RGB tuple.
    :param color: color code in hex string (e.g.: '#FF0000') or RGB tuple (e.g.: (255, 0, 0))
    :return: RGB tuple (R, G, B)
    """
    if isinstance(color, tuple):
        if len(color) != 3:
            raise ValueError("RGB tuple must have 3 elements.")
        rgb = (int(color[0]), int(color[1]), int(color[2]))
        return rgb
    if isinstance(color, str):
        if not color.startswith("#") or len(color) != 7:
            raise ValueError("Hex color code must be in the format '#RRGGBB'.")
        color = color.removeprefix("#")
        rgb = tuple(int(color[i: i + 2], 16) for i in (0, 2, 4))
        return rgb
    raise ValueError("Invalid color format. Must be hex string or RGB tuple.")


def make_empty_figure(
        subtitles: Union[str, Sequence[str]],
        sharex=False,
        sharey=False,
        subplots_vspace: float = 0.1,
        subplots_hspace: float = 0.1,
) -> Tuple[go.Figure, int, int]:
    if isinstance(subtitles, str):
        subtitles = [subtitles]
    ncols = 1 if len(subtitles) <= 3 else 2 if len(subtitles) <= 8 else 3
    nrows = len(subtitles) if len(subtitles) <= 3 else sum(divmod(len(subtitles), ncols))
    fig = make_subplots(
        rows=nrows, cols=ncols,
        shared_xaxes=sharex, shared_yaxes=sharey,
        vertical_spacing=subplots_vspace, horizontal_spacing=subplots_hspace,
        subplot_titles=subtitles,
    )
    return fig, nrows, ncols



from abc import ABC
from typing import Sequence, Dict, Union, Callable

from peyes._DataModels.Event import BaseEvent, EventSequenceType

OneToOneEventMatchesType = Dict[BaseEvent, BaseEvent]
OneToManyEventMatchesType = Dict[BaseEvent, EventSequenceType]
EventMatchesType = Union[OneToOneEventMatchesType, OneToManyEventMatchesType]
EventMatchingFunctionType = Callable[[EventSequenceType, EventSequenceType], EventMatchesType]


class EventMatcher(ABC):
    """
    Implementation of different methods to match two sequences of gaze-events, that may have been detected by different
    human annotators or detection algorithms, as discussed in section "Event Matching Methods" in the article:
        Startsev, M., Zemblys, R. Evaluating Eye Movement Event Detection: A Review of the State of the Art
        Behav Res 55, 1653–1714 (2023). https://doi.org/10.3758/s13428-021-01763-7
    """

    @staticmethod
    def generic_matching(
            ground_truth: EventSequenceType,
            predictions: EventSequenceType,
            allow_cross_matching: bool,
            min_overlap: float = 0,
            min_iou: float = - float("inf"),
            max_l2_timing_offset: float = float("inf"),
            max_onset_difference: float = float("inf"),
            max_offset_difference: float = float("inf"),
            reduction: str = "all"
    ) -> EventMatchesType:
        """
        Match each ground-truth event to a predicted event(s) that satisfies the specified criteria.

        :param ground_truth: sequence of ground-truth events
        :param predictions: sequence of predicted events
        :param allow_cross_matching: if True, a ground-truth event can match a predicted event of a different type
        :param min_overlap: minimum overlap time (normalized to GT duration) to consider a possible match (between 0 and 1)
        :param min_iou: minimum intersection-over-union to consider a possible match
        :param max_l2_timing_offset: maximum L2-timing-offset to consider a possible match
        :param max_onset_difference: maximum absolute difference (in ms) between the start times of the GT and predicted events
        :param max_offset_difference: maximum absolute difference (in ms) between the end times of the GT and predicted events
        :param reduction: name of reduction function used to choose a predicted event(s) from multiple matching ones:
            - 'all': return all matched events
            - 'first': return the first matched event
            - 'last': return the last matched event
            - 'longest': return the longest matched event
            - 'max overlap': return the matched event with maximum overlap with the GT event (normalized by GT duration)
            - 'iou': return the matched event with the maximum intersection-over-union with the GT event
            - 'l2': return the matched event with the minimum L2-timing-offset with the GT event
            - 'onset difference': return the matched event with the least onset-time difference
            - 'offset difference': return the matched event with the least offset-time difference
        :return: dictionary, where keys are ground-truth events and values are their matched predicted event(s)
        :raises NotImplementedError: if the reduction function is not implemented
        """
        reduction = reduction.lower().replace("_", " ").replace("-", " ").strip()
        matches = {}
        matched_predictions = set()
        for gt in ground_truth:
            unmatched_predictions = [p for p in predictions if p not in matched_predictions]
            possible_matches = EventMatcher.__find_matches(
                gt=gt,
                predictions=unmatched_predictions,
                allow_cross_matching=allow_cross_matching,
                min_overlap=min_overlap,
                min_iou=min_iou,
                max_l2_timing_offset=max_l2_timing_offset,
                max_onset_latency=max_onset_difference,
                max_offset_latency=max_offset_difference
            )
            p = EventMatcher.__choose_match(gt, possible_matches, reduction)
            if len(p):
                matches[gt] = p
            if reduction != "all":
                # If reduction is not 'all', cannot allow multiple matches for the same prediction
                matched_predictions.update(p)

        # verify output integrity
        if reduction != "all":
            assert all(len(v) == 1 for v in matches.values()), "Multiple matches for a GT event"
            matches = {k: v[0] for k, v in matches.items()}
            assert len(matches.values()) == len(set(matches.values())), "Matched predictions are not unique"
        return matches

    @staticmethod
    def first_overlap(
            ground_truth: EventSequenceType,
            predictions: EventSequenceType,
            min_overlap: float = 0,
            allow_cross_matching: bool = True
    ) -> OneToOneEventMatchesType:
        """
        Matches the first predicted event that overlaps with each ground-truth event, above a minimal overlap time.
        """
        return EventMatcher.generic_matching(
            ground_truth, predictions, allow_cross_matching, min_overlap=min_overlap, reduction="first"
        )

    @staticmethod
    def last_overlap(
            ground_truth: EventSequenceType,
            predictions: EventSequenceType,
            min_overlap: float = 0,
            allow_cross_matching: bool = True
    ) -> Dict[BaseEvent, BaseEvent]:
        """
        Matches the last predicted event that overlaps with each ground-truth event, above a minimal overlap time.
        """
        return EventMatcher.generic_matching(
            ground_truth, predictions, allow_cross_matching, min_overlap=min_overlap, reduction="last"
        )

    @staticmethod
    def max_overlap(
            ground_truth: EventSequenceType,
            predictions: EventSequenceType,
            min_overlap: float = 0,
            allow_cross_matching: bool = True
    ) -> OneToOneEventMatchesType:
        """
        Matches the predicted event with maximum overlap with each ground-truth event, above a minimal overlap time.
        Overlap-time is normalized by the duration of the ground-truth event, so values are between 0 and 1.
        """
        return EventMatcher.generic_matching(
            ground_truth, predictions, allow_cross_matching, min_overlap=min_overlap, reduction="max overlap"
        )

    @staticmethod
    def longest_overlapping_event(
            ground_truth: EventSequenceType,
            predictions: EventSequenceType,
            min_overlap: float = 0,
            allow_cross_matching: bool = True
    ) -> OneToOneEventMatchesType:
        """
        Matches the longest predicted event that overlaps with each ground-truth event, above a minimal overlap time.
        """
        return EventMatcher.generic_matching(
            ground_truth, predictions, allow_cross_matching, min_overlap=min_overlap, reduction="longest"
        )

    @staticmethod
    def iou(
            ground_truth: EventSequenceType,
            predictions: EventSequenceType,
            min_iou: float = 0,
            allow_cross_matching: bool = True
    ) -> OneToOneEventMatchesType:
        """
        Matches the predicted event with maximum intersection-over-union with each ground-truth event, above a minimal value.
        """
        return EventMatcher.generic_matching(
            ground_truth, predictions, allow_cross_matching, min_iou=min_iou, reduction="iou"
        )

    @staticmethod
    def l2_timing(
            ground_truth: EventSequenceType,
            predictions: EventSequenceType,
            max_l2: float = 0,
            allow_cross_matching: bool = True
    ) -> OneToOneEventMatchesType:
        """
        Matches the predicted event with minimum L2-timing-offset with each ground-truth event, below a maximum l2 value.
        """
        return EventMatcher.generic_matching(
            ground_truth, predictions, allow_cross_matching, max_l2_timing_offset=max_l2, reduction="l2"
        )

    @staticmethod
    def onset_difference(
            ground_truth: EventSequenceType,
            predictions: EventSequenceType,
            max_onset_difference: float = 0,
            allow_cross_matching: bool = True
    ) -> OneToOneEventMatchesType:
        """
        Matches the predicted event with least onset difference with each ground-truth event, below a maximum latency.
        """
        return EventMatcher.generic_matching(
            ground_truth, predictions, allow_cross_matching,
            max_onset_difference=max_onset_difference, reduction="onset difference"
        )

    @staticmethod
    def offset_difference(
            ground_truth: EventSequenceType,
            predictions: EventSequenceType,
            max_offset_difference: float = 0,
            allow_cross_matching: bool = True
    ) -> OneToOneEventMatchesType:
        """
        Matches the predicted event with least offset latency with each ground-truth event, below a maximum latency.
        """
        return EventMatcher.generic_matching(
            ground_truth, predictions, allow_cross_matching,
            max_offset_difference=max_offset_difference, reduction="offset difference"
        )

    @staticmethod
    def window_based(
            ground_truth: EventSequenceType,
            predictions: EventSequenceType,
            max_onset_difference: float = 0,
            max_offset_difference: float = 0,
            allow_cross_matching: bool = True,
            reduction: str = "onset difference"
    ) -> OneToOneEventMatchesType:
        """
        Finds all predicted events with onset- and offset-latencies within a specified window for each ground-truth event,
        and chooses the best gt-prediction match based on the specified reduction function.
        """
        return EventMatcher.generic_matching(
            ground_truth, predictions, allow_cross_matching,
            max_onset_difference=max_onset_difference, max_offset_difference=max_offset_difference,
            reduction=reduction
        )

    @staticmethod
    def __find_matches(
            gt: BaseEvent,
            predictions: EventSequenceType,
            allow_cross_matching: bool,
            min_overlap: float,
            min_iou: float,
            max_l2_timing_offset: float,
            max_onset_latency: float,
            max_offset_latency: float
    ) -> EventSequenceType:
        """
        Find predicted events that are possible matches for the ground-truth event, based on the specified criteria.

        :param gt: ground-truth event
        :param predictions: sequence of predicted events
        :param allow_cross_matching: if True, a GT event can match a predicted event of a different type
        :param min_overlap: minimum overlap time to consider a possible match
        :param min_iou: minimum intersection-over-union to consider a possible match
        :param max_l2_timing_offset: maximum L2-timing-offset to consider a possible match
        :param max_onset_latency: maximum absolute difference between the start times of the GT and predicted events
        :param max_offset_latency: maximum absolute difference between the end times of the GT and predicted events
        :return: sequence of predicted events that are possible matches for the ground-truth event
        """
        if not allow_cross_matching:
            predictions = [p for p in predictions if p.label == gt.label]
        predictions = [p for p in predictions if
                       gt.time_overlap(p) >= min_overlap and
                       gt.time_iou(p) >= min_iou and
                       gt.time_l2(p) <= max_l2_timing_offset and
                       abs(p.start_time - gt.start_time) <= max_onset_latency and
                       abs(p.end_time - gt.end_time) <= max_offset_latency]
        return predictions

    @staticmethod
    def __choose_match(gt: BaseEvent, matches: EventSequenceType, reduction: str) -> EventSequenceType:
        """
        Choose predicted event(s) matching the ground-truth event, based on the reduction function.
        Possible reduction functions:
            - 'all': return all matched events
            - 'first': return the first matched event
            - 'last': return the last matched event
            - 'longest': return the longest matched event
            - 'max overlap': return the matched event with maximum overlap with the GT event (normalized by GT duration)
            - 'iou': return the matched event with the maximum intersection-over-union with the GT event
            - 'l2': return the matched event with the minimum L2-timing-offset with the GT event
            - 'onset difference': return the matched event with the least onset difference
            - 'offset difference': return the matched event with the least offset difference

        :param gt: ground-truth event
        :param matches: sequence of predicted events matching with the GT event
        :param reduction: reduction function to choose a predicted event from multiple matching ones

        :return: predicted event(s) matching the ground-truth event
        :raises NotImplementedError: if the reduction function is not implemented
        """
        reduction = reduction.lower().replace("_", " ").replace("-", " ").strip()
        if len(matches) == 0:
            return []
        if len(matches) == 1:
            return [matches[0]]
        if reduction == "all":
            return matches
        if reduction == "first":
            return [min(matches, key=lambda e: e.start_time)]
        if reduction == "last":
            return [max(matches, key=lambda e: e.start_time)]
        if reduction == "longest":
            return [max(matches, key=lambda e: e.duration)]
        if reduction == "max overlap":
            return [max(matches, key=lambda e: gt.time_overlap(e))]
        if reduction == "iou":
            return [max(matches, key=lambda e: gt.time_iou(e))]
        if reduction in {"l2", "timing l2", "l2 timing offset"}:
            return [min(matches, key=lambda e: gt.time_l2(e))]
        if reduction == "onset difference":
            return [min(matches, key=lambda e: abs(e.start_time - gt.start_time))]
        if reduction == "offset difference":
            return [min(matches, key=lambda e: abs(e.end_time - gt.end_time))]
        raise NotImplementedError(f"Reduction function '{reduction}' is not implemented")



import os
from typing import Optional, Union, Sequence, List, Dict

import numpy as np
import pandas as pd
import plotly.express as px

import peyes
from peyes._DataModels.UnparsedEventLabel import UnparsedEventLabelType, UnparsedEventLabelSequenceType

###########################

BASE_DIR = r'S:\Lab-Shared\Experiments\pEYES'   # lab-shared directory
DATASETS_DIR = os.path.join(BASE_DIR, "datasets")
OUTPUT_DIR = os.path.join(BASE_DIR, peyes.constants.OUTPUT_STR)

###########################

GLOBAL_STR = "global"
SDT_STR = "sdt"
GT_STR, PRED_STR = "gt", "pred"
MATCHING_SCHEME_STR = "matching_scheme"

###########################

COLORMAP_TYPE = Union[Dict[str, str], Sequence[str]]
DEFAULT_DISCRETE_COLORMAP = px.colors.qualitative.Dark24
DEFAULT_CONTINUOUS_COLORMAP = px.colors.sequential.Viridis

###########################

DATASET_ANNOTATORS = {
    "lund2013": ["RA", "MN"],
    "irf": ['RZ'],
    "hfc": ['IH', 'DN', 'JV', 'JB', 'JF', 'KH', 'MN', 'MS', 'PZ', 'RA', 'RH', 'TC']
}
_ALL_ANNOTATORS = [annot for annotators in DATASET_ANNOTATORS.values() for annot in annotators]

METRICS_CONFIG = {
    # metric -> (name, order, value range)

    peyes.constants.ACCURACY_STR: ("Accuracy", 1, [0, 1]),
    peyes.constants.ONSET_STR: ("Onset", 1, None),
    f"{peyes.constants.AMPLITUDE_STR}_{peyes.constants.DIFFERENCE_STR}": ("Amplitude Difference", 1, None),

    peyes.constants.BALANCED_ACCURACY_STR: ("Balanced Accuracy", 2, [0, 1]),
    peyes.constants.OFFSET_STR: ("Offset", 2, None),
    f"{peyes.constants.AZIMUTH_STR}_{peyes.constants.DIFFERENCE_STR}": ("Azimuth Difference", 2, None),
    peyes.constants.MATCH_RATIO_STR: ("Match Ratio", 2, [0, 1]),

    peyes.constants.COHENS_KAPPA_STR: ("Cohen's Kappa", 3, [-1, 1]),
    f"center_{peyes.constants.PIXEL_STR}_{peyes.constants.DISTANCE_STR}": ("Center Distance", 3, None),

    peyes.constants.MCC_STR: ("MCC", 4, [-1, 1]),
    f"{peyes.constants.ONSET_STR}_{peyes.constants.DIFFERENCE_STR}": ("Onset Difference", 4, [-15, 15]),

    peyes.constants.RECALL_STR: ("Recall", 5, [0, 1]),
    f"{peyes.constants.OFFSET_STR}_{peyes.constants.DIFFERENCE_STR}": ("Offset Difference", 5, [-15, 15]),

    peyes.constants.PRECISION_STR: ("Precision", 6, [0, 1]),
    f"{peyes.constants.DURATION_STR}_{peyes.constants.DIFFERENCE_STR}": ("Duration Difference", 6, [-30, 30]),

    peyes.constants.F1_STR: ("f1", 7, [0, 1]),
    f"{peyes.constants.TIME_STR}_overlap": ("Time Overlap", 7, [0, 1]),

    peyes.constants.COMPLEMENT_NLD_STR: ("1 - NLD", 8, [0, 1]),
    f"{peyes.constants.TIME_STR}_iou": ("Time IoU", 8, [0, 1]),

    peyes.constants.FALSE_ALARM_RATE_STR: ("F.A. Rate", 8, [0, 1]),
    f"{peyes.constants.TIME_STR}_l2": ("Time L2", 8, None),

    peyes.constants.D_PRIME_STR: ("d'", 9, None),

    peyes.constants.CRITERION_STR: ("Criterion", 10, None),
}

###########################################


def load_dataset(dataset_name: str, verbose: bool = True) -> pd.DataFrame:
    if dataset_name == "lund2013":
        dataset = peyes.datasets.lund2013(directory=DATASETS_DIR, save=True, verbose=verbose)
    elif dataset_name == "irf":
        dataset = peyes.datasets.irf(directory=DATASETS_DIR, save=True, verbose=verbose)
    elif dataset_name == "hfc":
        dataset = peyes.datasets.hfc(directory=DATASETS_DIR, save=True, verbose=verbose)
    else:
        raise ValueError(f"Unknown dataset name: {dataset_name}")
    return dataset


def get_filename_for_labels(
        labels: Optional[Union[UnparsedEventLabelType, UnparsedEventLabelSequenceType]] = None,
        prefix: str = "",
        suffix: str = "",
        extension: str = "pkl"
) -> str:
    if prefix and not prefix.endswith("_"):
        prefix += "_"
    if suffix and not suffix.startswith("_"):
        suffix = "_" + suffix
    if labels is None:
        return f"{prefix}all_labels{suffix}.{extension}"
    elif isinstance(labels, UnparsedEventLabelType):
        return f"{prefix}{peyes.parse_label(labels).name.lower()}{suffix}.{extension}"
    elif isinstance(labels, Sequence) and all(isinstance(l, UnparsedEventLabelType) for l in labels):
        return f"{prefix}{'_'.join([peyes.parse_label(l).name.lower() for l in labels])}{suffix}.{extension}"
    else:
        raise TypeError(f"Unknown pos_labels type: {type(labels)}")


def get_trials_for_stimulus_type(
        dataset_name: str,
        stimulus_type: Union[str, List[str]],
) -> List[int]:
    if not stimulus_type:
        raise ValueError("Stimulus type is not specified")
    if isinstance(stimulus_type, str):
        stimulus_type = [stimulus_type]
    stimulus_type = list(set([stmtp.lower().strip() for stmtp in stimulus_type if isinstance(stmtp, str)]))
    dataset = load_dataset(dataset_name, verbose=False)
    is_stimulus_type = dataset[peyes.constants.STIMULUS_TYPE_STR].str.lower().isin(stimulus_type)
    trials = dataset.loc[is_stimulus_type, peyes.constants.TRIAL_ID_STR].unique()
    return trials


def get_labeler_color(labeler: str, idx: int, colors) -> str:
    colors = colors or DEFAULT_DISCRETE_COLORMAP
    if isinstance(colors, list):
        return colors[idx % len(colors)]
    elif isinstance(colors, dict):
        possibilities = [
            labeler, labeler.strip().lower(), labeler.strip().lower().removesuffix("detector"),
        ]
        for p in possibilities:
            if p in colors:
                return colors[p]
        return colors[idx % len(colors)]
    else:
        raise TypeError(f"Unknown colors type: {type(colors)}")


def sort_labelers(labelers: Sequence[str]) -> List[str]:
    labelers = list(set(labelers))
    return sorted(labelers, key=lambda l: _get_labeler_index(l, labelers))


def _get_labeler_index(labeler_name: str, detectors_names: Sequence[str]) -> float:
    if labeler_name in _ALL_ANNOTATORS:
        return _ALL_ANNOTATORS.index(labeler_name) / 10
    det_idx = detectors_names.index(labeler_name)
    if "ivt" in labeler_name.strip().lower():
        return det_idx + 10
    if "ivvt" in labeler_name.strip().lower():
        return det_idx + 20
    if "idt" in labeler_name.strip().lower():
        return det_idx + 30
    if "idvt" in labeler_name.strip().lower():
        return det_idx + 40
    if "engbert" in labeler_name.strip().lower():
        return det_idx + 50
    if "nh" in labeler_name.strip().lower():
        return det_idx + 60
    if "remodnav" in labeler_name.strip().lower():
        return det_idx + 70
    return det_idx + 100



import unittest

import numpy as np

import peyes._utils.constants as cnst
from peyes._DataModels.Event import FixationEvent, SaccadeEvent
from peyes._DataModels.EventLabelEnum import EventLabelEnum
from peyes._utils.pixel_utils import pixels_to_visual_angle


class TestEvent(unittest.TestCase):
    _PS, _VD = 1, 1

    def test_init(self):
        t = np.arange(20)
        x = np.sin(np.pi * t / 10)
        y = np.cos(np.pi * t / 10)
        self.assertRaises(AssertionError, FixationEvent, t=t, x=x, y=y[:-1])
        self.assertRaises(AssertionError, FixationEvent, t=t, x=x, y=y, pixel_size=-self._PS)
        self.assertRaises(AssertionError, FixationEvent, t=t, x=x, y=y, viewer_distance=-self._VD)
        f1 = FixationEvent(t=t)
        f2 = FixationEvent(t=t, x=None, y=None)
        f3 = FixationEvent(t=t, x=x, y=y, pixel_size=self._PS, viewer_distance=self._VD)
        self.assertTrue(f1 == f2)
        self.assertFalse(f1 == f3)
        self.assertEqual(str(f1), "FIXATION(19.00ms)")

    def test_make(self):
        # TODO
        self.assertTrue(True)

    def test_properties(self):
        fix_x, fix_y = np.full(21, 40), np.hstack([np.arange(50, 0, -5), np.arange(0, 51, 5)])
        t = np.arange(21)
        f = FixationEvent(t=t, x=fix_x, y=fix_y, pixel_size=self._PS, viewer_distance=self._VD)
        self.assertEqual(f.label, EventLabelEnum.FIXATION)
        self.assertEqual(f.duration, 20)
        self.assertEqual(f.center_pixel, (np.nanmean(fix_x), np.nanmean(fix_y)))
        self.assertEqual(f.pixel_std, (np.nanstd(fix_x), np.nanstd(fix_y)))
        self.assertEqual(f.distance, 0.0)
        self.assertEqual(f.amplitude, 0.0)
        self.assertEqual(f.azimuth, 0.0)
        self.assertEqual(f.cumulative_distance, 100)
        self.assertEqual(f.cumulative_amplitude, pixels_to_visual_angle(100, self._VD, self._PS))
        self.assertEqual(f.x_dispersion, 0.0)
        self.assertEqual(f.y_dispersion, pixels_to_visual_angle(50, self._VD, self._PS))
        self.assertEqual(f.dispersion, pixels_to_visual_angle(50, self._VD, self._PS))

    def test_velocity(self):
        t, x, y = np.arange(21), np.full(21, 40), np.hstack([np.arange(50, 0, -5), np.arange(0, 51, 5)])
        expected_px_vel = np.full_like(t, 5000, dtype=float)
        f = FixationEvent(t=t, x=x, y=y, pixel_size=self._PS, viewer_distance=self._VD)
        expected_px_vel[0] = np.nan  # first velocity is undefined
        self.assertTrue(np.allclose(f.velocities('px'), expected_px_vel, equal_nan=True))
        expected_deg_vel = np.vectorize(pixels_to_visual_angle)(expected_px_vel, self._VD, self._PS)
        self.assertTrue(np.allclose(f.velocities('deg'), expected_deg_vel, equal_nan=True))
        expected_rad_vel = np.vectorize(pixels_to_visual_angle)(expected_px_vel, self._VD, self._PS, use_radians=True)
        self.assertTrue(np.allclose(f.velocities('rad'), expected_rad_vel, equal_nan=True))
        self.assertRaises(ValueError, f.velocities, 'foobar')

    def test_overlaps(self):
        t, x, y = np.arange(21), np.full(21, 40), np.hstack([np.arange(50, 0, -5), np.arange(0, 51, 5)])
        f1 = FixationEvent(t=t, x=x, y=y, pixel_size=self._PS, viewer_distance=self._VD)
        f2 = FixationEvent(t=t+10, x=x, y=y, pixel_size=self._PS, viewer_distance=self._VD)
        self.assertFalse(f1 == f2)
        self.assertEqual(f1.time_overlap(f2, normalize=False), 10)
        self.assertEqual(f1.time_overlap(f2, normalize=True), 0.5)
        self.assertEqual(f1.time_iou(f2), 1/3)
        self.assertEqual(f1.time_l2(f2), 10 * np.sqrt(2))

    def test_duration_outliers(self):
        t = np.arange(21)
        s = SaccadeEvent(t=t)
        self.assertFalse(s.is_outlier)
        f = FixationEvent(t=t)
        self.assertTrue(f.is_outlier)
        self.assertEqual(f.get_outlier_reasons(), [cnst.MIN_DURATION_STR])



import os

import numpy as np
import peyes

import analysis.utils as u

DATASET_NAME = "lund2013"
STIMULUS_TYPE = peyes.constants.IMAGE_STR
GT1, GT2 = "RA", "MN"
MULTI_COMP = "fdr_bh"   # FDR Benjamini-Hochberg correction for multiple comparisons
ALPHA = 0.05

_DEFAULT_VALUES_STR = "default_values"
_FIGURES_STR = "figures"
PROCESSED_DATA_DIR = os.path.join(u.OUTPUT_DIR, _DEFAULT_VALUES_STR)
FIGURES_DIR = os.path.join(u.OUTPUT_DIR, _DEFAULT_VALUES_STR, DATASET_NAME, _FIGURES_STR)
os.makedirs(FIGURES_DIR, exist_ok=True)

_default_detector_params = dict(missing_value=np.nan, min_event_duration=4, pad_blinks_time=0)
DEFAULTS_DETECTORS = {
    "ivt": peyes.create_detector("ivt", **_default_detector_params),
    "ivvt": peyes.create_detector("ivvt", **_default_detector_params),
    "idt": peyes.create_detector("idt", **_default_detector_params),
    "idvt": peyes.create_detector("idvt", **_default_detector_params),
    "engbert": peyes.create_detector("engbert", **_default_detector_params),
    "nh": peyes.create_detector("nh", **_default_detector_params),
    "remodnav": peyes.create_detector("remodnav", show_warnings=False, **_default_detector_params),
}
